from uuid import uuid4

from django.contrib import admin, messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext

from edc_pylabels.models import LabelConfiguration

from ...models import Stock


@admin.action(description="Print labels")
def print_labels(modeladmin, request, queryset):
    selected = [str(obj.pk) for obj in queryset]
    if len(selected) > 0:
        session_uuid = str(uuid4())
        request.session[session_uuid] = selected
        url = reverse(
            "edc_pharmacy:print_labels_url",
            kwargs={"session_uuid": session_uuid, "model": "stock"},
        )
        return HttpResponseRedirect(url)
    return None


@admin.action(description="Print labels")
def print_labels_from_repack_request(modeladmin, request, queryset):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        session_uuid = str(uuid4())
        stock_qs = Stock.objects.values_list("pk", flat=True).filter(
            repack_request=queryset.first()
        )
        if stock_qs.exists():
            request.session[session_uuid] = [o for o in stock_qs]
            url = reverse(
                "edc_pharmacy:print_labels_url",
                kwargs={"session_uuid": session_uuid, "model": "stock"},
            )
            return HttpResponseRedirect(url)
    return None


@admin.action(description="Print labels")
def print_labels_from_receive(modeladmin, request, queryset):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        session_uuid = str(uuid4())
        stock_qs = Stock.objects.values_list("pk", flat=True).filter(
            receive_item__receive=queryset.first()
        )
        if stock_qs.exists():
            request.session[session_uuid] = [o for o in stock_qs]
            url = reverse(
                "edc_pharmacy:print_labels_url",
                kwargs={"session_uuid": session_uuid, "model": "stock"},
            )
            return HttpResponseRedirect(url)
    return None


@admin.action(description="Print labels")
def print_labels_from_receive_item(modeladmin, request, queryset):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        session_uuid = str(uuid4())
        stock_qs = Stock.objects.values_list("pk", flat=True).filter(
            receive_item=queryset.first()
        )
        if stock_qs.exists():
            request.session[session_uuid] = [o for o in stock_qs]
            url = reverse(
                "edc_pharmacy:print_labels_url",
                kwargs={"session_uuid": session_uuid, "model": "stock"},
            )
            return HttpResponseRedirect(url)
    return None


@admin.action(description="Print labels ordered by stock code")
def print_labels_from_stock_request_item(modeladmin, request, queryset):
    session_uuid = str(uuid4())
    try:
        stock_qs = Stock.objects.values_list("pk", flat=True).filter(
            code__in=[obj.allocation.stock.code for obj in queryset.all()]
        )
    except ObjectDoesNotExist:
        messages.add_message(
            request,
            messages.ERROR,
            "Allocate stock to subject before printing labels.",
        )
    else:
        if stock_qs.exists():
            request.session[session_uuid] = [o for o in stock_qs]
            try:
                label_configuration = LabelConfiguration.objects.get(
                    requires_allocation=True, name="patient_barcode"
                )
            except ObjectDoesNotExist:
                label_configuration = ""
            else:
                label_configuration = label_configuration.name
            url = reverse(
                "edc_pharmacy:print_labels_url",
                kwargs={
                    "session_uuid": session_uuid,
                    "model": "stock",
                    "label_configuration": label_configuration,
                },
            )
            return HttpResponseRedirect(url)
    return None


@admin.action(description="Print labels ordered by stock code")
def print_labels_from_stock_request_by_code(modeladmin, request, queryset):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext("Select one and only one item"),
        )
    else:
        session_uuid = str(uuid4())
        stock_request = queryset.first()
        stock_qs = Stock.objects.values_list("pk", flat=True).filter(
            code__in=[
                obj.allocation.stock.code
                for obj in stock_request.stockrequestitem_set.filter(
                    allocation__isnull=False,
                    allocation__stock__isnull=False,
                )
            ]
        )
        if stock_qs.exists():
            request.session[session_uuid] = [o for o in stock_qs.order_by("code")]
            try:
                label_configuration = LabelConfiguration.objects.get(
                    requires_allocation=True, name="patient_barcode"
                )
            except ObjectDoesNotExist:
                label_configuration = ""
            else:
                label_configuration = label_configuration.name
            url = reverse(
                "edc_pharmacy:print_labels_url",
                kwargs={
                    "session_uuid": session_uuid,
                    "label_configuration": label_configuration,
                    "model": "stock",
                },
            )
            return HttpResponseRedirect(url)
    return None



# Generated by Django 3.2 on 2021-05-10 17:36

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0018_auto_20210203_0706"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataquery",
            name="visit_code_sequence",
            field=models.IntegerField(
                blank=True,
                default=0,
                help_text="Defaults to '0'. For example, when combined with the visit code `1000` would make `1000.0`.",
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(25),
                ],
                verbose_name="Visit code sequence",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldataquery",
            name="visit_code_sequence",
            field=models.IntegerField(
                blank=True,
                default=0,
                help_text="Defaults to '0'. For example, when combined with the visit code `1000` would make `1000.0`.",
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(25),
                ],
                verbose_name="Visit code sequence",
            ),
        ),
    ]



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_form_runners"
    verbose_name = "Edc Form Runners"
    description = ""
    admin_site_name = "edc_form_runners_admin"
    include_in_administration_section = True
    has_exportable_data = True
    default_auto_field = "django.db.models.BigAutoField"



from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import admin
from django.contrib.admin import AdminSite as DjangoAdminSite
from django.contrib.sites.shortcuts import get_current_site
from django.template.response import TemplateResponse
from django.urls import NoReverseMatch, reverse

from edc_protocol.research_protocol_config import ResearchProtocolConfig

admin.site.enable_nav_sidebar = False


class EdcAdminSite(DjangoAdminSite):
    """
    Add to your project urls.py
        path("edc_action_item/", edc_action_item.urls),

    -OR-
    To include this in the administration section set
    `AppConfig.include_in_administration_section = True`
    in your apps.py. (See also View `edc_dashboard.administration.py`).

    If set to `include_in_administration_section=True`, add a local `urls.py`

        from django.urls.conf import path
        from django.views.generic import RedirectView

        app_name = "edc_action_item"

        urlpatterns = [
            path("", RedirectView.as_view(url="admin/"), name="home_url"),
        ]

    and then add to your project urls.py

        path("edc_action_item/admin/", edc_action_item_admin.urls),
        path("edc_action_item/", include("edc_action_item.urls")),

    """

    index_template = "edc_model_admin/admin/index.html"
    app_index_template = "edc_model_admin/admin/app_index.html"
    login_template = "edc_auth/login.html"
    logout_template = "edc_auth/login.html"
    enable_nav_sidebar = False
    final_catch_all_view = True
    site_url = "/administration/"

    def __init__(
        self,
        name="admin",
        app_label=None,
        keep_delete_action=None,
        enable_nav_sidebar=None,
    ):
        self.app_label = app_label
        if enable_nav_sidebar is not None:
            self.enable_nav_sidebar = enable_nav_sidebar
        super().__init__(name)
        if not keep_delete_action:
            del self._actions["delete_selected"]

    @property
    def app_url(self):
        try:
            app_url = reverse(f"{self.name}:index")
        except NoReverseMatch:
            app_url = "#"
        return app_url

    def each_context(self, request):
        context = super().each_context(request)
        context.update(
            site_title=self.get_edc_site_title(request),
            site_header=self.get_edc_site_header(request),
            global_site=get_current_site(request),
            protocol_name=ResearchProtocolConfig().protocol_name,
            live_system=settings.LIVE_SYSTEM,
            DEBUG=settings.DEBUG,
            app_url=self.app_url,
            verbose_name=django_apps.get_app_config(self.app_label).verbose_name,
        )
        return context

    def get_edc_site_title(self, request) -> str:
        verbose_name = django_apps.get_app_config(self.app_label).verbose_name
        return verbose_name.replace(
            ResearchProtocolConfig().project_name,
            f"{ResearchProtocolConfig().project_name} @ "
            f"{get_current_site(request).name.title()} ",
        )

    def get_edc_site_header(self, request) -> str:
        return self.get_edc_site_title(request)

    def get_edc_index_title(self, request) -> str:
        return self.get_edc_site_title(request)

    def index(self, request, extra_context=None):
        app_list = self.get_app_list(request)
        context = {
            **self.each_context(request),
            "subtitle": None,
            "app_list": app_list,
            **(extra_context or {}),
        }
        request.current_app = self.name
        return TemplateResponse(request, self.index_template or "admin/index.html", context)



from django.db import models

from ..model_mixin_factories import fbg_model_mixin_factory


class FbgModelMixin(fbg_model_mixin_factory("fbg"), models.Model):
    """A model mixin of fields for the FBG"""

    class Meta:
        abstract = True



class AppointmentStatusError(Exception):
    pass


class AppointmentBaselineError(Exception):
    pass


class AppointmentReasonUpdaterError(Exception):
    pass


class AppointmentReasonUpdaterCrfsExistsError(Exception):
    pass


class AppointmentReasonUpdaterRequisitionsExistsError(Exception):
    pass


class UnscheduledAppointmentError(Exception):
    pass


class AppointmentCreateError(Exception):
    pass


class AppointmentDatetimeError(Exception):
    pass


class UnknownVisitCode(Exception):
    pass


class AppointmentWindowError(Exception):
    pass


class AppointmentPermissionsRequired(Exception):
    pass


class AppointmentMissingValuesError(Exception):
    pass


class UnscheduledAppointmentNotAllowed(Exception):
    pass


class InvalidVisitCodeSequencesError(Exception):
    pass


class InvalidTimepointError(Exception):
    pass


class InvalidParentAppointmentStatusError(Exception):
    pass


class InvalidParentAppointmentMissingVisitError(Exception):
    pass


class AppointmentInProgressError(Exception):
    pass


class AppointmentCreatorError(Exception):
    pass


class CreateAppointmentError(Exception):
    pass


class NextAppointmentModelError(Exception):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_visit_tracking.utils import get_related_visit_model_cls

from .site import site_metadata_rules

if TYPE_CHECKING:
    related_visit_model_cls = get_related_visit_model_cls()


class MetadataRuleEvaluator:
    """Main class to evaluate rules.

    Used by model mixin.
    """

    def __init__(
        self,
        related_visit: related_visit_model_cls = None,
        app_label: str | None = None,
        allow_create: bool | None = None,
    ) -> None:
        self.related_visit = related_visit
        self.app_labels = [app_label] if app_label else []
        self.related_visit_model = related_visit._meta.label_lower
        self.allow_create = allow_create
        if not self.app_labels:
            for rule_groups in site_metadata_rules.registry.values():
                for rule_group in rule_groups:
                    if (
                        rule_group._meta.related_visit_model == self.related_visit_model
                        and rule_group._meta.app_label not in self.app_labels
                    ):
                        self.app_labels.append(rule_group._meta.app_label)

    def evaluate_rules(self) -> None:
        for app_label in self.app_labels:
            for rule_group in site_metadata_rules.registry.get(app_label, []):
                rule_group.evaluate_rules(
                    related_visit=self.related_visit, allow_create=self.allow_create
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.core.handlers.wsgi import WSGIRequest
from django.utils import timezone

from ..exceptions import ConfirmAtSiteError

if TYPE_CHECKING:
    from uuid import UUID

    from ..models import (
        ConfirmationAtSite,
        ConfirmationAtSiteItem,
        Location,
        Stock,
        StockTransfer,
    )


def confirm_stock_at_site(
    stock_transfer: StockTransfer,
    stock_codes: list[str],
    location: UUID,
    request: WSGIRequest | None = None,
) -> tuple[list[str], list[str], list[str]]:
    """Confirm stock instances given a list of stock codes
    and a request/receive pk.

    Called from ConfirmStock view.

    See also: confirm_stock_action
    """
    confirmed_by = request.user.username
    stock_model_cls: type[Stock] = django_apps.get_model("edc_pharmacy.stock")
    confirmation_at_site_model_cls: type[ConfirmationAtSite] = django_apps.get_model(
        "edc_pharmacy.confirmationatsite"
    )
    confirmation_at_site_item_model_cls: type[ConfirmationAtSiteItem] = django_apps.get_model(
        "edc_pharmacy.confirmationatsiteitem"
    )
    location_model_cls: type[Location] = django_apps.get_model("edc_pharmacy.location")

    location = location_model_cls.objects.get(pk=location)

    confirmation_at_site, _ = confirmation_at_site_model_cls.objects.get_or_create(
        stock_transfer=stock_transfer,
        location=location,
    )

    confirmed, already_confirmed, invalid = [], [], []
    stock_codes = [s.strip() for s in stock_codes]
    for stock_code in stock_codes:
        if (
            not stock_model_cls.objects.filter(code=stock_code).exists()
            or not stock_transfer.stocktransferitem_set.filter(stock__code=stock_code).exists()
        ):
            invalid.append(stock_code)
        else:
            try:
                stock = stock_model_cls.objects.get(
                    code=stock_code,
                    location=location,
                    confirmation__isnull=False,
                    allocation__isnull=False,
                    confirmationatsiteitem__isnull=True,
                )
            except ObjectDoesNotExist:
                already_confirmed.append(stock_code)
            else:
                obj = confirmation_at_site_item_model_cls(
                    confirmation_at_site=confirmation_at_site,
                    stock=stock,
                    confirmed_datetime=timezone.now(),
                    confirmed_by=confirmed_by,
                    user_created=confirmed_by,
                    created=timezone.now(),
                )
                try:
                    obj.save()
                except ConfirmAtSiteError as e:
                    messages.add_message(request, messages.ERROR, str(e))
                    invalid.append(stock_code)
                else:
                    confirmed.append(stock_code)
    return confirmed, already_confirmed, invalid


__all__ = ["confirm_stock_at_site"]



from django.apps import apps as django_apps

from .base_box_item_model_wrapper import BaseBoxItemModelWrapper

app_config = django_apps.get_app_config("edc_lab_dashboard")


class VerifyBoxItemModelWrapper(BaseBoxItemModelWrapper):
    next_url_name = "verify_box_listboard_url"
    action_name = "verify"



# Generated by Django 4.0.4 on 2022-05-19 19:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0026_auto_20220501_1548"),
    ]

    operations = [
        migrations.AddField(
            model_name="appointment",
            name="document_status",
            field=models.CharField(
                choices=[
                    ("INCOMPLETE", "Incomplete (some data pending)"),
                    ("COMPLETE", "Complete"),
                ],
                default="COMPLETE",
                help_text="If some data is still pending, flag as incomplete",
                max_length=25,
                verbose_name="Document status",
            ),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="document_status",
            field=models.CharField(
                choices=[
                    ("INCOMPLETE", "Incomplete (some data pending)"),
                    ("COMPLETE", "Complete"),
                ],
                default="COMPLETE",
                help_text="If some data is still pending, flag as incomplete",
                max_length=25,
                verbose_name="Document status",
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.urls.conf import re_path

from edc_constants.constants import UUID_PATTERN

from .url_names import url_names

if TYPE_CHECKING:
    from django.urls import URLPattern
    from django.views import View as BaseView

    from .view_mixins import UrlRequestContextMixin

    class View(UrlRequestContextMixin, BaseView): ...


class UrlConfig:
    def __init__(
        self,
        *,
        url_name: str,
        namespace: str,
        view_class: type[View | UrlRequestContextMixin],
        label: str,
        identifier_label: str,
        identifier_pattern: str,
    ):
        self.identifier_label = identifier_label
        self.identifier_pattern = identifier_pattern
        self.label = label
        self.url_name = url_name
        self.view_class = view_class

        # register {urlname, namespace:urlname} with url_names
        url_names.register(url=self.url_name, namespace=namespace)

    @property
    def dashboard_urls(self) -> list[URLPattern]:
        """Returns url patterns."""
        return [
            re_path(
                "{label}/"
                "(?P<{identifier_label}>{identifier_pattern})/"
                r"(?P<visit_schedule_name>\w+)/"
                r"(?P<schedule_name>\w+)/"
                r"(?P<visit_code>\w+)/"
                r"(?P<unscheduled>\w+)/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/"
                "(?P<{identifier_label}>{identifier_pattern})/"
                r"(?P<visit_schedule_name>\w+)/"
                r"(?P<schedule_name>\w+)/"
                r"(?P<visit_code>\w+)/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/"
                "(?P<{identifier_label}>{identifier_pattern})/"
                "(?P<appointment>{uuid_pattern})/"
                r"(?P<scanning>\d)/"
                r"(?P<error>\d)/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                        uuid_pattern=UUID_PATTERN.pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/"
                "(?P<{identifier_label}>{identifier_pattern})/"
                "(?P<appointment>{uuid_pattern})/"
                r"(?P<reason>\w+)/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                        uuid_pattern=UUID_PATTERN.pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/"
                "(?P<{identifier_label}>{identifier_pattern})/"
                "(?P<appointment>{uuid_pattern})/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                        uuid_pattern=UUID_PATTERN.pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/"
                "(?P<{identifier_label}>{identifier_pattern})/"
                r"(?P<schedule_name>\w+)/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/(?P<{identifier_label}>{identifier_pattern})/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
        ]

    @property
    def listboard_urls(self) -> list[URLPattern]:
        """Returns url patterns.

        configs = [(listboard_url, listboard_view_class, label), (), ...]
        """
        return [
            re_path(
                "{label}/(?P<{identifier_label}>{identifier_pattern})/"
                r"(?P<page>\d+)/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                "{label}/(?P<{identifier_label}>{identifier_pattern})/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                r"{label}/(?P<page>\d+)/".format(**dict(label=self.label)),
                self.view_class.as_view(),
                name=self.url_name,
            ),
            re_path(
                r"{label}/".format(**dict(label=self.label)),
                self.view_class.as_view(),
                name=self.url_name,
            ),
        ]

    @property
    def review_listboard_urls(self) -> list[URLPattern]:
        url_patterns = [
            re_path(
                "{label}/(?P<{identifier_label}>{identifier_pattern})/"
                "(?P<appointment>{uuid_pattern})/".format(
                    **dict(
                        label=self.label,
                        identifier_label=self.identifier_label,
                        identifier_pattern=self.identifier_pattern,
                        uuid_pattern=UUID_PATTERN.pattern,
                    )
                ),
                self.view_class.as_view(),
                name=self.url_name,
            )
        ]
        url_patterns.extend(self.listboard_urls)
        return url_patterns



from django.core.management.base import BaseCommand

from edc_appointment.constants import COMPLETE_APPT, INCOMPLETE_APPT
from edc_appointment.forms import AppointmentForm
from edc_appointment.models import Appointment
from edc_metadata.models import CrfMetadata


def close_appointments():
    for obj in Appointment.objects.filter(appt_status=INCOMPLETE_APPT).order_by(
        "subject_identifier", "visit_code", "visit_code_sequence"
    ):
        data = obj.__dict__
        data.update(appt_status=COMPLETE_APPT, appt_close_datetime=obj.modified)
        form = AppointmentForm(data=data, instance=obj)
        form.is_valid()
        try:
            form.save(commit=False)
        except ValueError:
            obj.refresh_from_db()
            print(
                obj.subject_identifier,
                obj.visit_code,
                obj.visit_code_sequence,
                obj.appt_status,
            )
        else:
            obj.refresh_from_db()
            obj.appt_status = COMPLETE_APPT
            obj.save()

    for obj in Appointment.objects.filter(appt_status=COMPLETE_APPT).order_by(
        "subject_identifier", "visit_code", "visit_code_sequence"
    ):
        data = obj.__dict__
        data.update(appt_status=INCOMPLETE_APPT, appt_close_datetime=obj.modified)
        form = AppointmentForm(data=data, instance=obj)
        form.is_valid()
        try:
            form.save(commit=False)
        except ValueError:
            pass
        else:
            obj.refresh_from_db()
            print(
                obj.subject_identifier,
                obj.visit_code,
                obj.visit_code_sequence,
                obj.appt_status,
                CrfMetadata,
            )


class Command(BaseCommand):
    def handle(self, *args, **options):
        close_appointments()



from django.db import models
from tqdm import tqdm

from .generate_slug import generate_slug


class SearchSlugManager(models.Manager):
    def update_search_slugs(self) -> None:
        qs = self.all()
        for obj in tqdm(qs, total=qs.count()):
            obj.slug = generate_slug(obj, obj.get_search_slug_fields()) or ""
            obj.save_base(update_fields=["slug"])


class SearchSlugModelMixin(models.Model):
    def get_search_slug_fields(self) -> tuple[str, ...]:
        return ()

    slug = models.CharField(
        max_length=250,
        default="",
        editable=False,
        db_index=True,
        help_text="Hold slug field values for quick search. Excludes encrypted fields",
    )

    def save(self, *args, **kwargs):
        self.slug = generate_slug(self, self.get_search_slug_fields()) or ""
        super().save(*args, **kwargs)

    class Meta:
        abstract = True
        indexes = (models.Index(fields=["slug"]),)



# Generated by Django 5.2.6 on 2025-09-24 04:56

import django_revision.revision_field
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_reportable", "0007_alter_gradingdata_age_phrase_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="gradingdata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="gradingexception",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalgradingdata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmolecularweight",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="historicalnormaldata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="molecularweight",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                default="",
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                verbose_name="Revision",
            ),
        ),
    ]



import csv
import os
from datetime import datetime

from django.core.exceptions import MultipleObjectsReturned
from django.core.management.base import BaseCommand, CommandError

from edc_constants.constants import CLOSED

from ...models import ExportPlan, ExportReceipt, ExportTransaction


class Command(BaseCommand):
    args = "<receipt filename>"
    help = "Import a receipt file for recently exported transactions."
    option_list = BaseCommand.option_list

    def handle(self, *args, **options):
        header = []
        has_rejects = False
        try:
            self.ack_filename = args[0]
            _, app_label1, app_label2, self.object_name, timestamp = (
                self.ack_filename.split("/").pop().split("_")
            )
            self.app_label = app_label1 + "_" + app_label2
            self.timestamp, self.extension = timestamp.split(".")
        except IndexError:
            raise CommandError("Usage: import_receipts <receipt filename>")
        print("reading file...")
        error_filepath = os.path.join(
            os.path.expanduser(self.export_plan.target_path) or "", self.error_filename
        )
        with open(self.ack_filename) as f, open(error_filepath, "w") as error_file:
            rows = csv.reader(f, delimiter="|")
            writer = csv.writer(error_file, delimiter="|")
            for row in rows:
                if not header:
                    header = row
                    writer.writerow(header)
                    continue
                try:
                    export_uuid = row[header.index("export_UUID")]
                except ValueError as e:
                    writer.writerow(f"error reading file. Got {e}")
                    print(f"Failed to process file {self.ack_filename}")
                    raise ValueError(e)
                if ExportTransaction.objects.filter(export_uuid=export_uuid):
                    for export_transaction in ExportTransaction.objects.filter(
                        export_uuid=export_uuid
                    ):
                        self.update_or_create_export_transaction(export_transaction)
                        print("  accepted: " + export_uuid)

                else:
                    has_rejects = True
                    writer.writerow(row)
                    print("  rejected: " + export_uuid)
        self.clean_up(has_rejects)

    def clean_up(self, has_rejects):
        if has_rejects:
            target_path = os.path.join(os.path.expanduser(self.export_plan.target_path))
            print("Some receipts were rejected.")
            print(f"See file {self.error_filename} in {target_path}")
        else:
            os.remove(self.error_filename)
            print("Success")

    @property
    def error_filename(self):
        try:
            app_label1, app_label2 = self.app_labe1.split("_")
            error_filename = (
                "_".join(["error", app_label1, app_label2, self.object_name, self.timestamp])
                + "."
                + self.extension
            )
        except ValueError:
            CommandError(
                "Invalid file name. Expected format xxx_app_label_objectname_timestamp.xxx. "
                f"Got {self.ack_filename}"
            )
        return error_filename

    @property
    def export_plan(self):
        try:
            export_plan = ExportPlan.objects.get(
                app_label=self.app_label, object_name=self.object_name
            )
        except ExportPlan.DoesNotExist as e:
            CommandError(
                f"ExportPlan not found for {self.app_label}, {self.object_name}. "
                "Check filename format or create an ExportPlan. "
                f"Got {e}"
            )
        return export_plan

    def update_or_create_export_transaction(self, export_transaction):
        try:
            ExportReceipt.objects.get(export_uuid=export_transaction.export_uuid)
        except MultipleObjectsReturned:
            pass
        except ExportReceipt.DoesNotExist:
            ExportReceipt.objects.create(
                export_uuid=export_transaction.export_uuid,
                app_label=export_transaction.app_label,
                object_name=export_transaction.object_name,
                timestamp=self.timestamp,
                received_datetime=datetime.today(),
                tx_pk=export_transaction.tx_pk,
            )
        export_transaction.status = CLOSED
        export_transaction.received = True
        export_transaction.received_datetime = datetime.today()
        export_transaction.save()



from __future__ import annotations

from typing import TYPE_CHECKING

from .grading_exception_model_cls import grading_exception_model_cls

if TYPE_CHECKING:
    from ..models import ReferenceRangeCollection

__all__ = ["update_grading_exceptions"]


def update_grading_exceptions(
    reference_range_collection: ReferenceRangeCollection,
    reportable_grades_exceptions: dict[str, list[str]] | None = None,
    keep_existing: bool | None = None,
):
    reportable_grades_exceptions = reportable_grades_exceptions or {}
    if not keep_existing:
        grading_exception_model_cls().objects.filter(
            reference_range_collection=reference_range_collection
        ).delete()
    for label, grades in reportable_grades_exceptions.items():
        grades = [int(g) for g in grades]
        grading_exception_model_cls().objects.get_or_create(
            reference_range_collection=reference_range_collection,
            label=label,
            grade1=True if 1 in grades else False,
            grade2=True if 2 in grades else False,
            grade3=True if 3 in grades else False,
            grade4=True if 4 in grades else False,
        )



# Generated by Django 6.0 on 2025-03-12 14:01

from django.db import migrations

import edc_pharmacy.models.stock.allocation


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0062_auto_20250312_1433"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="allocation",
            managers=[
                ("objects", edc_pharmacy.models.stock.allocation.Manager()),
            ],
        ),
        migrations.RemoveField(
            model_name="allocation",
            name="site",
        ),
        migrations.RemoveField(
            model_name="historicalallocation",
            name="site",
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-25 20:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "0011_update_proxy_permissions"),
        ("edc_auth", "0010_auto_20191022_0422"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserRole",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=25)),
                ("groups", models.ManyToManyField(to="auth.Group")),
            ],
        ),
        migrations.AddField(
            model_name="userprofile",
            name="user_roles",
            field=models.ManyToManyField(to="edc_auth.UserRole"),
        ),
    ]



from django.apps import apps as django_apps

from .timepoint import Timepoint


class TimepointDoesNotExist(Exception):  # noqa: N818
    pass


class TimepointConfigError(Exception):
    pass


class TimepointCollection:
    """Contains a collection of Timepoint instances."""

    def __init__(self, timepoints=None):
        self._timepoints = {}
        for timepoint in timepoints:
            self.add(**timepoint.__dict__)

    def __iter__(self):
        return iter(self._timepoints)

    def add(self, model=None, **kwargs):
        self._timepoints.update({model: Timepoint(model=model, **kwargs)})

    def get_model(self, model=None):
        """Returns the timepoint's model class."""
        timepoint = self.get(model)
        return django_apps.get_model(timepoint.model)

    def get(self, model=None):
        """Returns the timepoint class for this model."""
        try:
            timepoint = self._timepoints[model]
        except KeyError as e:
            raise TimepointDoesNotExist(
                f"No timepoint has been configured with {model}. "
                "See AppConfig for edc_timepoint. Hint: Perhaps you are using a custom "
                "`Appointment` model?"
            ) from e
        return timepoint



from django import forms
from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist

from edc_form_validators import FormValidator

from ..constants import LTFU
from ..utils import get_ltfu_model_name


class LtfuFormValidatorMixin(FormValidator):
    ltfu_model = get_ltfu_model_name()
    offschedule_reason_field = "offschedule_reason"

    @property
    def ltfu_model_cls(self):
        return django_apps.get_model(self.ltfu_model)

    def validate_ltfu(self):
        subject_identifier = (
            self.cleaned_data.get("subject_identifier") or self.instance.subject_identifier
        )

        try:
            self.ltfu_model_cls.objects.get(subject_identifier=subject_identifier)
        except ObjectDoesNotExist:
            if self.offschedule_reason_field not in self.cleaned_data:
                raise ImproperlyConfigured(
                    "Unknown offschedule_reason_field. "
                    f"Got '{self.offschedule_reason_field}'. "
                    f"See form {self.__class__.__name__}"
                )
            if self.cleaned_data.get(self.offschedule_reason_field) == LTFU:
                raise forms.ValidationError(
                    {
                        self.offschedule_reason_field: (
                            "Patient was lost to followup, please complete "
                            f"'{self.ltfu_model_cls._meta.verbose_name}' "
                            "form first."
                        )
                    }
                )



class ColumnMapError(Exception):
    pass


class ColumnMap:
    column_names = None
    mappings = {}

    def __init__(self, dataframe=None):
        self.dataframe = dataframe
        for column_name in self.column_names:
            if column_name in list(self.dataframe.columns):
                self._map(column_name)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.column_names})"

    def _map(self, column_name):
        mapping = self.mappings.get(column_name)
        if not mapping:
            raise ColumnMapError(f"Mapping not found for column {column_name}")
        self.dataframe[column_name] = self.dataframe[column_name].map(mapping.get)



from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist


def get_related_or_none(obj, attr):
    try:
        return getattr(obj, attr)
    except ObjectDoesNotExist:
        return None



# Generated by Django 5.1 on 2024-08-15 23:56

from django.db import migrations
from django.db.migrations import RunPython

from edc_constants.constants import FEEDBACK, PENDING


def update_status(apps, schema_editor):
    for model_cls in apps.get_models():
        if getattr(model_cls, "status", None):
            print(
                f"Updated `{model_cls._meta.label_lower}.status` from `{FEEDBACK}` "
                f"to `{PENDING}`."
            )
            model_cls.objects.filter(status=FEEDBACK).update(status=PENDING)


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0016_alter_note_options_alter_qareportlog_options"),
    ]

    operations = [RunPython(update_status)]



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

name = "edc_subject_dashboard.middleware.DashboardMiddleware"
if name not in settings.MIDDLEWARE:
    raise ImproperlyConfigured(f"Missing middleware. Expected {name}.")



from django.db import models


class SubjectOnScheduleModelMixin(models.Model):
    """A model mixin for a consent or other model
    that when saved updates a subject to be `on schedule`.

    Used when the "Onschedule" model is not a user model.

    This is NOT for the "Onschedule" model. See `OnScheduleModelMixin`.
    """

    def put_subject_on_schedule_on_post_save(self, created):
        """A wrapper to put a subject on schedule.

        Called in signals.
        """
        if created:
            self.schedule.put_on_schedule(
                subject_identifier=self.subject_identifier,
                onschedule_datetime=self.consent_datetime,
            )

    class Meta:
        abstract = True



# Generated by Django 5.2.1 on 2025-06-04 16:31

import uuid

import _socket
import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("edc_reportable", "0003_referencerangecollection_grade1_and_more"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name="referencerangecollection",
            name="grade3",
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name="referencerangecollection",
            name="grade4",
            field=models.BooleanField(default=True),
        ),
        migrations.CreateModel(
            name="HistoricalGradingData",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("label", models.CharField(max_length=25)),
                ("description", models.CharField(max_length=255, null=True)),
                ("reference_group", models.CharField(max_length=25, null=True)),
                ("lower", models.FloatField(null=True)),
                ("lower_operator", models.CharField(max_length=15, null=True)),
                ("lower_inclusive", models.BooleanField(default=False)),
                ("lln", models.CharField(default=None, max_length=15, null=True)),
                ("upper", models.FloatField(null=True)),
                ("upper_operator", models.CharField(max_length=2, null=True)),
                ("upper_inclusive", models.BooleanField(default=False)),
                ("uln", models.CharField(default=None, max_length=15, null=True)),
                (
                    "gender",
                    models.CharField(
                        max_length=1,
                        validators=[django.core.validators.RegexValidator("[MF]{1}")],
                    ),
                ),
                ("units", models.CharField(max_length=15)),
                ("age_units", models.CharField(max_length=15)),
                ("age_lower", models.IntegerField()),
                ("age_lower_operator", models.CharField(max_length=15)),
                ("age_lower_inclusive", models.BooleanField(default=False)),
                ("age_upper", models.IntegerField(null=True)),
                ("age_upper_operator", models.CharField(max_length=15, null=True)),
                ("age_upper_inclusive", models.BooleanField(null=True)),
                ("fasting", models.BooleanField(default=False)),
                (
                    "phrase",
                    models.CharField(
                        help_text="calculated by the formula instance",
                        max_length=50,
                        null=True,
                        verbose_name="Value phrase",
                    ),
                ),
                (
                    "age_phrase",
                    models.CharField(
                        help_text="calculated in save()", max_length=25, null=True
                    ),
                ),
                ("grade", models.IntegerField()),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "reference_range_collection",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_reportable.referencerangecollection",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Grading Reference",
                "verbose_name_plural": "historical Grading References",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalNormalData",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("label", models.CharField(max_length=25)),
                ("description", models.CharField(max_length=255, null=True)),
                ("reference_group", models.CharField(max_length=25, null=True)),
                ("lower", models.FloatField(null=True)),
                ("lower_operator", models.CharField(max_length=15, null=True)),
                ("lower_inclusive", models.BooleanField(default=False)),
                ("lln", models.CharField(default=None, max_length=15, null=True)),
                ("upper", models.FloatField(null=True)),
                ("upper_operator", models.CharField(max_length=2, null=True)),
                ("upper_inclusive", models.BooleanField(default=False)),
                ("uln", models.CharField(default=None, max_length=15, null=True)),
                (
                    "gender",
                    models.CharField(
                        max_length=1,
                        validators=[django.core.validators.RegexValidator("[MF]{1}")],
                    ),
                ),
                ("units", models.CharField(max_length=15)),
                ("age_units", models.CharField(max_length=15)),
                ("age_lower", models.IntegerField()),
                ("age_lower_operator", models.CharField(max_length=15)),
                ("age_lower_inclusive", models.BooleanField(default=False)),
                ("age_upper", models.IntegerField(null=True)),
                ("age_upper_operator", models.CharField(max_length=15, null=True)),
                ("age_upper_inclusive", models.BooleanField(null=True)),
                ("fasting", models.BooleanField(default=False)),
                (
                    "phrase",
                    models.CharField(
                        help_text="calculated by the formula instance",
                        max_length=50,
                        null=True,
                        verbose_name="Value phrase",
                    ),
                ),
                (
                    "age_phrase",
                    models.CharField(
                        help_text="calculated in save()", max_length=25, null=True
                    ),
                ),
                (
                    "grade",
                    models.IntegerField(
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(1),
                            django.core.validators.MaxValueValidator(5),
                        ],
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField(db_index=True)),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "reference_range_collection",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_reportable.referencerangecollection",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Normal Reference",
                "verbose_name_plural": "historical Normal References",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": ("history_date", "history_id"),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



from __future__ import annotations

from datetime import date

from django.db import models

from edc_constants.choices import YES_NO
from edc_constants.constants import NOT_APPLICABLE, YES
from edc_dx import Diagnoses
from edc_model.models import DurationYMDField
from edc_model.utils import get_report_datetime_field_name
from edc_model.validators import date_not_future

from .calculate_date import update_calculated_date


class InitialReviewModelError(Exception):
    pass


def dx_initial_review_methods_model_mixin_factory():
    class AbstractModel(models.Model):
        fld_prefix: str | None = None

        def save(self, *args, **kwargs):
            if not self.diagnoses.get_dx_by_model(self) == YES:
                raise InitialReviewModelError(
                    "No diagnosis has been recorded. See clinical review. "
                    "Perhaps catch this in the form."
                )
            update_calculated_date(
                self,
                fld_prefix=self.fld_prefix,
                reference_field=get_report_datetime_field_name(),
            )
            super().save(*args, **kwargs)

        @property
        def diagnoses(self):
            subject_identifier = self.subject_identifier
            return Diagnoses(
                subject_identifier=subject_identifier,
                report_datetime=getattr(self, get_report_datetime_field_name()),
                lte=True,
            )

        def get_best_dx_date(self) -> date:
            return getattr(self, f"{self.fld_prefix}_date") or getattr(
                self, f"{self.fld_prefix}_calculated_date"
            )

        class Meta:
            abstract = True

    return AbstractModel


def dx_initial_review_model_mixin_factory(fld_prefix: str | None = None):
    fld_prefix = fld_prefix or "dx"

    class AbstractModel(dx_initial_review_methods_model_mixin_factory(), models.Model):
        class Meta:
            abstract = True

    opts = {
        "fld_prefix": fld_prefix,
        f"{fld_prefix}_date": models.DateField(
            verbose_name="Date patient diagnosed",
            null=True,
            blank=True,
            validators=[date_not_future],
            help_text="If possible, provide the exact date here instead of estimating.",
        ),
        f"{fld_prefix}_ago": DurationYMDField(
            verbose_name="If date not known, how long ago was the patient diagnosed?",
            null=True,
            blank=True,
            help_text="If possible, provide the exact date above instead of estimating here.",
        ),
        f"{fld_prefix}_calculated_date": models.DateField(
            verbose_name="Estimated diagnosis date",
            null=True,
            help_text=f"Calculated based on response to `{fld_prefix}dx_ago`",
            editable=False,
        ),
        f"{fld_prefix}_date_is_estimated": models.CharField(
            verbose_name="Was the diagnosis date estimated?",
            max_length=15,
            choices=YES_NO,
            default=NOT_APPLICABLE,
            editable=False,
        ),
    }

    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel



import re

from sqlalchemy import TextClause, text


class MysqlDialectError(Exception):
    pass


class MysqlDialect:
    def __init__(self, dbname=None):
        self.dbname = dbname

    def __repr__(self):
        return f"{self.__class__.__name__}({self.dbname})"

    @staticmethod
    def show_databases() -> tuple[TextClause, None]:
        sql = text("SELECT SCHEMA_NAME AS 'database' FROM INFORMATION_SCHEMA.SCHEMATA")
        return sql, None

    def show_tables(self, app_label: str = None) -> tuple[TextClause, dict]:
        params = {"dbname": self.dbname}
        select = "SELECT table_name FROM information_schema.tables"
        where = ["table_schema=:dbname"]
        if app_label:
            where.append("table_name LIKE :app_label ")
            params.update({"app_label": f"{app_label}%"})
        sql = text(f'{select} WHERE {" AND ".join(where)}')
        return sql, params

    def show_tables_with_columns(
        self, app_label: str = None, column_names: list[str] = None
    ) -> tuple[TextClause, dict]:
        column_names = "','".join(column_names)
        params = {
            "dbname": self.dbname,
            "app_label": f"{app_label}%",
            "column_names": column_names,
        }
        sql = text(
            "SELECT DISTINCT table_name FROM information_schema.columns "
            "WHERE table_schema=:dbname "
            "AND table_name LIKE :app_label "
            "AND column_name IN (:column_names)"
        )
        return sql, params

    def show_tables_without_columns(
        self, app_label: str = None, column_names: list[str] = None
    ) -> tuple[TextClause, dict]:
        column_names = "','".join(column_names)
        params = {
            "dbname": self.dbname,
            "app_label": f"{app_label}%",
            "column_names": column_names,
        }
        sql = text(
            "SELECT DISTINCT table_name FROM information_schema.tables as T "
            "WHERE T.table_schema = :dbname "
            "AND T.table_type = 'BASE TABLE' "
            "AND T.table_name LIKE :app_label "
            "AND NOT EXISTS ("
            "SELECT * FROM INFORMATION_SCHEMA.COLUMNS C "
            "WHERE C.table_schema = T.table_schema "
            "AND C.table_name = T.table_name "
            "AND C.column_name IN (:column_names)"
        )
        return sql, params

    @staticmethod
    def select_table(table_name: str = None) -> tuple[TextClause, dict]:
        if not re.match(r"^[A-Za-z0-9]+(?:_[A-Za-z0-9]+)*$", table_name):
            raise MysqlDialectError(f"Invalid table_name. Got {table_name}")
        params = {}
        sql = text(f"select * from {table_name}")  # nosec B608
        return sql, params

    @staticmethod
    def show_inline_tables(
        referenced_table_name: str = None,
    ) -> tuple[TextClause, dict]:
        params = {"referenced_table_name": referenced_table_name}
        sql = text(
            "SELECT DISTINCT referenced_table_name, table_name, "
            "column_name, referenced_column_name "
            "FROM information_schema.key_column_usage "
            "where referenced_table_name=:referenced_table_name"
        )
        return sql, params



from django.utils.translation import gettext_lazy as _

from edc_constants.constants import (
    CLINIC,
    ESTIMATED,
    HOME,
    HOSPITAL,
    HOSPITAL_NOTES,
    NOT_APPLICABLE,
    OTHER,
    OUTPATIENT_CARDS,
    PATIENT,
    PATIENT_REPRESENTATIVE,
    TELEPHONE,
)

list_data = {
    "edc_appointment.appointmenttype": [
        (CLINIC, _("In clinic")),
        (HOME, _("At home")),
        (HOSPITAL, _("In hospital")),
        (TELEPHONE, _("By telephone")),
        (NOT_APPLICABLE, _("Not applicable")),
    ],
    "edc_appointment.infosources": [
        (PATIENT, _("Patient")),
        (
            PATIENT_REPRESENTATIVE,
            _("Patient representative (e.g., next of kin, relative, guardian)"),
        ),
        (HOSPITAL_NOTES, _("Hospital notes")),
        (OUTPATIENT_CARDS, _("Outpatient cards")),
        (ESTIMATED, _("Estimated by research staff")),
        (OTHER, _("Other")),
    ],
}



from __future__ import annotations

import sys

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from ..single_site import SingleSite
from .get_or_create_site_obj import get_or_create_site_obj
from .get_or_create_site_profile_obj import get_or_create_site_profile_obj


class UpdateDjangoSitesError(Exception):
    pass


def get_sites():
    from ..site import sites  # prevent circular import  # noqa: PLC0415

    return sites


def add_or_update_django_sites(
    apps=None,
    single_sites: list[SingleSite] | tuple[SingleSite] | None = None,
    verbose: bool | None = None,
):
    """Removes default site and adds/updates given `sites`, etc.

    Title is stored in SiteProfile.

    kwargs:
        * sites: format
            sites = (
                (<site_id>, <site_name>, <title>),
                ...)
    """
    if verbose:
        sys.stdout.write("  * updating sites.\n")
    apps = apps or django_apps
    site_model_cls = apps.get_model("sites", "Site")
    try:
        obj = site_model_cls.objects.get(id=1)
    except ObjectDoesNotExist:
        pass
    else:
        # Delete will fail if you have an unmanaged model with an FK
        # to Site and `on_delete` is something other than DO_NOTHING.
        # See the comment in edc_appconfig.apps about why we might
        # unregister `create_default_site` post_migrate signal.
        obj.delete()
    if not single_sites:
        single_sites = get_sites().all().values()
    if not single_sites:
        raise UpdateDjangoSitesError("No sites have been registered.")
    for single_site in single_sites:
        if single_site.name == "edc_sites.sites":
            continue
        if verbose:
            sys.stdout.write(f"  * SingleSite: {single_site.site_id}: {single_site.domain}.\n")
        site_obj = get_or_create_site_obj(single_site, apps)
        if verbose:
            sys.stdout.write(f"    - Site model: {site_obj.id}: {site_obj.domain}.\n")
        get_or_create_site_profile_obj(single_site, site_obj, apps)
    return single_sites



from __future__ import annotations

import contextlib
import warnings
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q

from edc_registration.utils import get_registered_subject_model_cls

from .constants import (
    DEFAULT_ASSIGNMENT_DESCRIPTION_MAP,
    DEFAULT_ASSIGNMENT_MAP,
    RANDOMIZED,
)
from .exceptions import (
    AllocationError,
    AlreadyRandomized,
    InvalidAssignmentDescriptionMap,
    RandomizationError,
    RandomizationListAlreadyImported,
    RandomizationListFileNotFound,
)
from .randomization_list_importer import RandomizationListImporter
from .utils import get_randomization_list_path

if TYPE_CHECKING:
    from edc_registration.models import RegisteredSubject


__all__ = ["Randomizer"]


class Randomizer:
    """Selects and uses the next available slot in model
    RandomizationList (cls.model) for this site. A slot is used
    when the subject identifier is not None.

    This is the default randomizer class and is registered with
    `site_randomizer` by default. To prevent registration set
    settings.EDC_RANDOMIZATION_REGISTER_DEFAULT_RANDOMIZER=False.

    assignment_map: {<assignment:str>: <allocation:int>, ...}
    assignment_description_map: {<assignment:str>: <description:str>, ...}


    Usage:
        Randomizer(
            subject_identifier=subject_identifier,
            report_datetime=report_datetime,
            site=site,
            user=user,
            **kwargs,
        ).randomize()

    It is better to access this class via the site_randomizer through a signal
    on something like the subject_consent:

        site_randomizers.randomize(
            "default",
            subject_identifier=instance.subject_identifier,
            report_datetime=instance.consent_datetime,
            site=instance.site,
            user=instance.user_created,
            gender=instance.gender)



    """

    name: str = "default"
    model: str = "edc_randomization.randomizationlist"
    assignment_map: dict[str, int] = getattr(
        settings, "EDC_RANDOMIZATION_ASSIGNMENT_MAP", DEFAULT_ASSIGNMENT_MAP
    )
    assignment_description_map: dict[str, str] = getattr(
        settings,
        "EDC_RANDOMIZATION_ASSIGNMENT_DESCRIPTION_MAP",
        DEFAULT_ASSIGNMENT_DESCRIPTION_MAP,
    )
    filename: str = "randomization_list.csv"
    randomizationlist_folder: Path | str = get_randomization_list_path()
    extra_csv_fieldnames: tuple[str] | None = None
    trial_is_blinded: bool = True
    importer_cls: Any = RandomizationListImporter
    apps = None  # if not using django_apps

    def __init__(
        self,
        identifier: str | None = None,
        subject_identifier: str | None = None,
        identifier_attr: str | None = None,
        identifier_object_name: str | None = None,
        report_datetime: datetime | None = None,
        site: Any | None = None,
        user: str | None = None,
        **kwargs,  # noqa: ARG002
    ):
        self._model_obj = None
        self._registration_obj = None
        self.identifier_attr = identifier_attr or "subject_identifier"
        self.identifier_object_name = identifier_object_name or "subject"
        setattr(self, self.identifier_attr, identifier or subject_identifier)
        self.allocated_datetime = report_datetime
        self.site = site
        self.user = user
        self.validate_assignment_description_map()
        self.import_list(overwrite=False)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name},{self.randomizationlist_folder})"

    def __str__(self):
        return f"<{self.name} for file {self.randomizationlist_folder}>"

    @classmethod
    def get_assignment(cls, subject_identifier: str) -> str | None:
        try:
            obj = cls.model_cls().objects.get(subject_identifier=subject_identifier)
        except ObjectDoesNotExist:
            pass
        else:
            return obj.assignment
        return None

    def randomize(self):
        """Randomize a subject.

        Will raise RandomizationError if general problems;
        Will raise AlreadyRandomized if already randomized.
        """
        self.raise_if_already_randomized()
        required_instance_attrs = dict(
            allocated_datetime=self.allocated_datetime,
            user=self.user,
            site=self.site,
            **self.identifier_opts,
            **self.extra_required_instance_attrs,
        )

        if not all(required_instance_attrs.values()):
            raise RandomizationError(
                f"Randomization failed. Insufficient data. Got {required_instance_attrs}."
            )
        setattr(
            self.model_obj,
            self.identifier_attr,
            getattr(self, self.identifier_attr),
        )
        self.model_obj.allocated_datetime = self.allocated_datetime
        self.model_obj.allocated_user = self.user
        self.model_obj.allocated_site = self.site
        self.model_obj.allocated = True
        self.model_obj.save()
        # requery
        self._model_obj = self.model_cls().objects.get(
            allocated=True,
            allocated_datetime=self.allocated_datetime,
            **self.identifier_opts,
        )
        self.registration_obj.sid = self.sid
        self.registration_obj.randomization_datetime = self.model_obj.allocated_datetime
        self.registration_obj.registration_status = RANDOMIZED
        self.registration_obj.randomization_list_model = self.model_obj._meta.label_lower
        self.registration_obj.save()
        # requery
        self._registration_obj = self.get_registration_model_cls().objects.get(
            sid=self.model_obj.sid, **self.identifier_opts
        )

    @property
    def identifier_opts(self) -> dict[str, str]:
        return {self.identifier_attr: getattr(self, self.identifier_attr)}

    @classmethod
    def get_registration_model_cls(cls) -> type[RegisteredSubject]:
        return get_registered_subject_model_cls()

    @property
    def extra_required_instance_attrs(self):
        """Returns a dict of extra attributes that must have
        value on self.
        """
        return {}

    @property
    def sid(self):
        """Returns the SID."""
        if not self.model_obj.sid:
            raise RandomizationError(
                f"SID cannot be None. See {self.model_obj}. Got {self.model_obj.sid}"
            )
        return self.model_obj.sid

    @property
    def extra_model_obj_options(self):
        """Returns a dict of extra key/value pair for filtering the
        "rando" model.
        """
        return {}

    @classmethod
    def model_cls(cls):
        return (cls.apps or django_apps).get_model(cls.model)

    @property
    def model_obj(self):
        """Returns a "rando" model instance by selecting
        the next available SID.

        (e.g. RandomizationList)
        """
        if not self._model_obj:
            try:
                obj = self.model_cls().objects.get(**self.identifier_opts)
            except ObjectDoesNotExist as e:
                opts = dict(site_name=self.site.name, **self.extra_model_obj_options)
                self._model_obj = (
                    self.model_cls()
                    .objects.filter(**{f"{self.identifier_attr}__isnull": True}, **opts)
                    .order_by("sid")
                    .first()
                )
                if not self._model_obj:
                    fld_str = ", ".join([f"{k}=`{v}`" for k, v in opts.items()])
                    raise AllocationError(
                        f"Randomization failed. No additional SIDs available for {fld_str}."
                    ) from e
            else:
                raise AlreadyRandomized(
                    f"{self.identifier_object_name.title()} already randomized. "
                    f"Got {getattr(obj, self.identifier_attr)} SID={obj.sid}. "
                    f"Something is wrong. Are "
                    f"{self.get_registration_model_cls()._meta.label_lower} and "
                    f"{self.model_cls()._meta.label_lower} out of sync?.",
                    code=self.model_cls()._meta.label_lower,
                )
        return self._model_obj

    def raise_if_already_randomized(self) -> Any:
        """Forces a query, will raise if already randomized."""
        return self.registration_obj

    def validate_assignment_description_map(self) -> None:
        """Raises an exception if the assignment description map
        has extra or missing keys.

        Compares with the assignment map.
        """
        if sorted(list(self.assignment_map.keys())) != sorted(
            list(self.assignment_description_map.keys())
        ):
            raise InvalidAssignmentDescriptionMap(
                f"Invalid assignment description. See randomizer {self.name}. "
                f"Got {self.assignment_description_map}."
            )

    def get_unallocated_registration_obj(self):
        """Returns an unallocated registration model instance
        or raises.

        Called by `registration_obj`.
        """
        return self.get_registration_model_cls().objects.get(
            (Q(sid__isnull=True) | Q(sid="")), **self.identifier_opts
        )

    @property
    def registration_obj(self):
        """Returns an unrandomized instance of the registration model
        for this identifier or raises.

        By default, if SID is null, the instance has not been randomized.

        (e.g. RegisteredSubject).
        """

        if not self._registration_obj:
            try:
                self._registration_obj = self.get_unallocated_registration_obj()
            except ObjectDoesNotExist as e:
                try:
                    obj = self.get_registration_model_cls().objects.get(**self.identifier_opts)
                except ObjectDoesNotExist as e:
                    raise RandomizationError(
                        f"{self.identifier_object_name.title()} does not exist. "
                        f"Got {getattr(self, self.identifier_attr)}"
                    ) from e
                else:
                    raise AlreadyRandomized(
                        f"{self.identifier_object_name.title()} already randomized. "
                        f"See {self.get_registration_model_cls()._meta.verbose_name}. "
                        f"Got {getattr(obj, self.identifier_attr)} "
                        f"SID={obj.sid}",
                        code=self.get_registration_model_cls()._meta.label_lower,
                    ) from e
        return self._registration_obj

    @property
    def registered_subject(self):
        warnings.warn(
            "This property is decrecated in favor of `registration_obj`.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.registration_obj

    @classmethod
    def get_extra_list_display(cls) -> tuple[tuple[int, str], ...]:
        """Returns a list of tuples of (pos, field name) for ModelAdmin."""
        return ()

    @classmethod
    def get_extra_list_filter(cls) -> tuple[tuple[int, str], ...]:
        """Returns a list of tuples of (pos, field name) for ModelAdmin."""
        return cls.get_extra_list_display()

    @classmethod
    def get_randomizationlist_path(cls) -> Path:
        return Path(cls.randomizationlist_folder) / cls.filename

    @classmethod
    def import_list(cls, **kwargs) -> tuple[int, str]:
        result = (0, "")
        if not cls.get_randomizationlist_path().exists():
            raise RandomizationListFileNotFound(
                "Randomization list file not found. "
                f"Got `{cls.get_randomizationlist_path()}`. See Randomizer {cls.name}."
            )
        with contextlib.suppress(RandomizationListAlreadyImported):
            result = cls.importer_cls(
                assignment_map=cls.assignment_map,
                randomizationlist_path=cls.get_randomizationlist_path(),
                randomizer_model_cls=cls.model_cls(),
                randomizer_name=cls.name,
                extra_csv_fieldnames=cls.extra_csv_fieldnames,
                **kwargs,
            ).import_list(**kwargs)
        return result

    @classmethod
    def verify_list(cls, **kwargs) -> list[str]:
        return cls.importer_cls.verifier_cls(
            assignment_map=cls.assignment_map,
            randomizationlist_path=cls.get_randomizationlist_path(),
            randomizer_model_cls=cls.model_cls(),
            randomizer_name=cls.name,
            **kwargs,
        ).messages



from edc_pharmacy.models import Stock


def stock_unit_qty_in_check():
    """Check for any buckets where unit_qty_in < unit_qty_out

    Assumes 128 tabs
    """
    for obj in Stock.objects.filter(container__container_type__name="bucket"):
        stock_code = obj.code
        stock_unit_qty_in = obj.unit_qty_in
        stock_unit_qty_out = obj.unit_qty_out
        from_stock_count = Stock.objects.filter(from_stock__code=obj.code).count() * 128

        if not stock_unit_qty_in >= stock_unit_qty_out:
            print(stock_code, stock_unit_qty_in, stock_unit_qty_out, from_stock_count)
            print(f"**Error stock_code={stock_code}")



# Generated by Django 5.1.2 on 2024-11-03 17:09

import _socket
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("django_pylabels", "0003_rename_updated_labelspecification_modified"),
        ("edc_pylabels", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="label",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Label",
                "verbose_name_plural": "Labels",
            },
        ),
        migrations.AddField(
            model_name="label",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name="label",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AddField(
            model_name="label",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AddField(
            model_name="label",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AddField(
            model_name="label",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AddField(
            model_name="label",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="label",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="label",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name="label",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. Git repository tag:branch:commit.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
        migrations.AddField(
            model_name="label",
            name="user_created",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user created",
            ),
        ),
        migrations.AddField(
            model_name="label",
            name="user_modified",
            field=django_audit_fields.fields.userfield.UserField(
                blank=True,
                help_text="Updated by admin.save_model",
                max_length=50,
                verbose_name="user modified",
            ),
        ),
        migrations.AlterField(
            model_name="label",
            name="id",
            field=django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                blank=True,
                editable=False,
                help_text="System auto field. UUID primary key.",
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AddIndex(
            model_name="label",
            index=models.Index(
                fields=["modified", "created"], name="edc_pylabel_modifie_01500a_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="label",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_pylabel_user_mo_cf698a_idx",
            ),
        ),
    ]



from __future__ import annotations

from datetime import date
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import QuerySet
from multisite.exceptions import MultisiteSiteDoesNotExist

from edc_sites.site import sites as site_sites
from edc_sites.utils import get_site_model_cls
from edc_utils.date import to_local

from .exceptions import FacilityCountryError, FacilitySiteError, HolidayError
from .holidays_disabled import holidays_disabled

if TYPE_CHECKING:
    from django.contrib.sites.models import Site


class Holidays:
    """A class used by Facility to get holidays for the
    country of facility.
    """

    model: str = "edc_facility.holiday"

    def __init__(self, site: Site = None) -> None:
        self._holidays = None
        self.model_cls = django_apps.get_model(self.model)
        self._site: Site | None = site

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(country={self.country}, "
            f"time_zone={settings.TIME_ZONE})"
        )

    def __len__(self):
        return self.holidays.count()

    @property
    def local_dates(self) -> list[date]:
        return [obj.local_date for obj in self.holidays]

    @property
    def site(self) -> Site | None:
        """Returns the Site model instance."""
        if not self._site:
            try:
                self._site = get_site_model_cls().objects.get_current()
            except ObjectDoesNotExist as e:
                raise FacilitySiteError(
                    f"Unable to determine site. Cannot lookup holidays. "
                    f"settings.SITE_ID={settings.SITE_ID}. Got {e}"
                )
            except MultisiteSiteDoesNotExist as e:
                raise FacilitySiteError(
                    f"Unable to determine site. Cannot lookup holidays. "
                    f"settings.SITE_ID={settings.SITE_ID}. Got MultisiteSiteDoesNotExist({e})."
                )
        return self._site

    @property
    def country(self) -> str:
        """Returns country string.

        Requires SiteProfile from edc_sites to be updated.
        """
        country = site_sites.get(self.site.id).country
        if not country:
            raise FacilityCountryError("Unable to determine country.")
        return country

    @property
    def holidays(self) -> QuerySet:
        """Returns a holiday model instance for this country."""
        if not self._holidays:
            if holidays_disabled():
                self._holidays = self.model_cls.objects.none()
            elif not self.model_cls.objects.filter(country=self.country).exists():
                raise HolidayError(f"No holidays found for '{self.country}. See {self.model}.")
            self._holidays = self.model_cls.objects.filter(country=self.country)
        return self._holidays

    def is_holiday(self, utc_datetime=None) -> bool:
        """Returns True if the UTC datetime is a holiday."""
        local_date = to_local(utc_datetime).date()
        try:
            self.model_cls.objects.get(country=self.country, local_date=local_date)
        except ObjectDoesNotExist:
            return False
        return True



from celery import shared_task
from django.conf import settings

from edc_data_manager.models import QueryRule
from edc_data_manager.rule import RuleRunner

DATA_MANAGER_ENABLED = getattr(settings, "DATA_MANAGER_ENABLED", True)


@shared_task(name="update_query_rules")
def update_query_rules(pks=None):
    if DATA_MANAGER_ENABLED:
        if pks:
            query_rules = QueryRule.objects.filter(id__in=pks, active=True)
        else:
            query_rules = QueryRule.objects.filter(active=True)
        total_created = 0
        total_resolved = 0
        for query_rule_obj in query_rules:
            rule_runner = RuleRunner(query_rule_obj)
            created, resolved = rule_runner.run()
            total_created += created
            total_resolved += resolved
        return {"created": total_created, "resolved": total_resolved}
    return {"created": 0, "resolved": 0}



# Generated by Django 4.2.7 on 2023-11-28 02:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_metadata",
            "0024_rename_current_entry_title_crfmetadata_model_verbose_name_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="crfmetadata",
            name="document_user",
            field=models.CharField(max_length=50, null=True, verbose_name="User"),
        ),
        migrations.AddField(
            model_name="requisitionmetadata",
            name="document_user",
            field=models.CharField(max_length=50, null=True, verbose_name="User"),
        ),
        migrations.AlterField(
            model_name="crfmetadata",
            name="model_verbose_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Document"),
        ),
        migrations.AlterField(
            model_name="requisitionmetadata",
            name="model_verbose_name",
            field=models.CharField(max_length=250, null=True, verbose_name="Document"),
        ),
    ]



from typing import Any

from edc_constants.constants import NOT_APPLICABLE, NULL_STRING

from .base_form_validator import (
    APPLICABLE_ERROR,
    NOT_APPLICABLE_ERROR,
    BaseFormValidator,
)


class ApplicableFieldValidator(BaseFormValidator):
    def raise_applicable(
        self,
        field,
        msg: str | None = None,
        applicable_msg: str | None = None,
    ) -> None:
        message = {field: applicable_msg or f"This field is applicable. {msg or ''}".strip()}
        self.raise_validation_error(message, APPLICABLE_ERROR)

    def raise_not_applicable(
        self,
        field,
        msg: str | None = None,
        not_applicable_msg: str | None = None,
    ) -> None:
        message = {
            field: not_applicable_msg or f"This field is not applicable. {msg or ''}".strip()
        }
        self.raise_validation_error(message, NOT_APPLICABLE_ERROR)

    def applicable_if(
        self,
        *responses: Any,
        field: str,
        field_applicable: str,
        inverse: bool | None = None,
        is_instance_field: bool | None = None,
        msg: str | None = None,
        applicable_msg: str | None = None,
        not_applicable_msg: str | None = None,
        not_applicable_value=None,
    ) -> bool:
        return self.applicable(
            *responses,
            field=field,
            field_applicable=field_applicable,
            inverse=inverse,
            is_instance_field=is_instance_field,
            msg=msg,
            applicable_msg=applicable_msg,
            not_applicable_msg=not_applicable_msg,
            not_applicable_value=not_applicable_value,
        )

    def not_applicable_if(
        self,
        *responses: Any,
        field: str = None,
        field_applicable: str = None,
        inverse: bool | None = None,
        is_instance_field: bool | None = None,
        msg: str | None = None,
        applicable_msg: str | None = None,
        not_applicable_msg: str | None = None,
        not_applicable_value=None,
    ) -> bool:
        return self.not_applicable(
            *responses,
            field=field,
            field_applicable=field_applicable,
            inverse=inverse,
            is_instance_field=is_instance_field,
            msg=msg,
            applicable_msg=applicable_msg,
            not_applicable_msg=not_applicable_msg,
            not_applicable_value=not_applicable_value,
        )

    def not_applicable_only_if(
        self, *responses, field=None, field_applicable=None, is_instance_field=None
    ) -> bool:
        if is_instance_field:
            self.update_cleaned_data_from_instance(field)
        field_value = self.get(field)
        field_applicable_value = self.get(field_applicable)

        if field_value in responses and (
            field_applicable_value and field_applicable_value is not None
        ):
            self.raise_not_applicable(field_applicable)
        return False

    def applicable(
        self,
        *responses: Any,
        field: str,
        field_applicable: str,
        inverse: bool | None = None,
        is_instance_field: bool | None = None,
        msg: str | None = None,
        applicable_msg: str | None = None,
        not_applicable_msg: str | None = None,
        not_applicable_value=None,
    ) -> bool:
        """Returns False or raises a validation error for field
        pattern where response to question 1 makes
        question 2 applicable.
        """
        inverse = True if inverse is None else inverse
        not_applicable = not_applicable_value or NOT_APPLICABLE
        if is_instance_field:
            self.update_cleaned_data_from_instance(field)
        if field in self.cleaned_data and field_applicable in self.cleaned_data:
            field_value = self.get(field)
            field_applicable_value = self.get(field_applicable)

            if field_value in responses and (
                field_applicable_value in (None, NULL_STRING, not_applicable)
            ):
                self.raise_applicable(field_applicable, msg=msg, applicable_msg=applicable_msg)
            elif (
                field_value not in responses
                and field_applicable_value != not_applicable
                and inverse
            ):
                self.raise_not_applicable(
                    field_applicable, msg=msg, not_applicable_msg=not_applicable_msg
                )
        return False

    def not_applicable(
        self,
        *responses: Any,
        field: str = None,
        field_applicable: str = None,
        inverse: bool | None = None,
        is_instance_field: bool | None = None,
        msg: str | None = None,
        applicable_msg: str | None = None,
        not_applicable_msg: str | None = None,
        not_applicable_value=None,
    ) -> bool:
        """Returns False or raises a validation error for field
        pattern where response to question 1 makes
        question 2 NOT applicable.
        """
        inverse = True if inverse is None else inverse
        not_applicable = not_applicable_value or NOT_APPLICABLE
        if is_instance_field:
            self.update_cleaned_data_from_instance(field)
        if field in self.cleaned_data and field_applicable in self.cleaned_data:
            if self.get(field) in responses and self.get(field_applicable) != not_applicable:
                self.raise_not_applicable(
                    field_applicable, msg=msg, not_applicable_msg=not_applicable_msg
                )
            elif inverse and (
                self.get(field) not in responses
                and self.get(field_applicable) == not_applicable
            ):
                self.raise_applicable(field_applicable, msg=msg, applicable_msg=applicable_msg)
        return False

    def applicable_if_true(
        self,
        condition: bool,
        field_applicable: str = None,
        applicable_msg: str | None = None,
        not_applicable_msg: str | None = None,
        inverse: bool | None = None,
        not_applicable_value=None,
    ) -> bool:
        inverse = True if inverse is None else inverse
        not_applicable = not_applicable_value or NOT_APPLICABLE
        if field_applicable in self.cleaned_data:
            if condition and self.get(field_applicable) == not_applicable:
                self.raise_applicable(field_applicable, msg=applicable_msg)
            elif not condition and self.get(field_applicable) != not_applicable:
                if inverse:
                    self.raise_not_applicable(
                        field_applicable,
                        not_applicable_msg=not_applicable_msg,
                        msg=not_applicable_msg,
                    )
        return False

    def not_applicable_if_true(
        self,
        condition: bool,
        field_applicable: str = None,
        applicable_msg: str | None = None,
        not_applicable_msg: str | None = None,
        inverse: bool | None = None,
        not_applicable_value=None,
    ) -> bool:
        inverse = True if inverse is None else inverse
        not_applicable = not_applicable_value or NOT_APPLICABLE
        if field_applicable in self.cleaned_data:
            if condition and self.get(field_applicable) != not_applicable:
                self.raise_not_applicable(field_applicable, msg=not_applicable_msg)
            elif not condition and self.get(field_applicable) == not_applicable:
                if inverse:
                    self.raise_applicable(
                        field_applicable,
                        applicable_msg=applicable_msg,
                        msg=applicable_msg,
                    )
        return False



from __future__ import annotations

from ..database import Database


def get_table_names(
    app_label: str,
    with_columns: list[str] | None = None,
    without_columns: list[str] | None = None,
    db_cls: type[Database] | None = None,
) -> list[str]:
    """Returns a list of table names for this app_label."""
    db = (db_cls or Database)()

    if with_columns:
        df = db.show_tables_with_columns(app_label, with_columns)
    elif without_columns:
        df = db.show_tables_without_columns(app_label, without_columns)
    else:
        df = db.show_tables(app_label)
    df = df.rename(columns={"TABLE_NAME": "table_name"})
    return list(df.table_name)


def get_model_names(
    app_label: str,
    with_columns: list[str] | None = None,
    without_columns: list[str] | None = None,
    db_cls: type[Database] | None = None,
    exclude_historical: bool | None = None,
    exclude_views: bool | None = None,
) -> list[str]:
    """Returns a list of model names derived from the table name
    for this app_label.

    Will return the wrong model name if the db_table is unmanaged
    and naming does not follow Django convention. `exclude_views`
    is False by default for this reason.

    TODO: why not use django_apps.get_models()? Is it because
          of encrypted fields?
    """
    model_names = []
    exclude_views = True if exclude_views is None else exclude_views
    for table_name in get_table_names(
        app_label,
        with_columns=with_columns,
        without_columns=without_columns,
        db_cls=db_cls,
    ):
        model_name = table_name.split(app_label)[1][1::]
        if (exclude_historical and model_name.startswith("historical")) or (
            exclude_views and model_name.endswith("_view")
        ):
            continue
        model_names.append(f"{app_label}.{model_name}")
    return model_names



import csv
import sys
from itertools import islice
from pathlib import Path

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style
from django.db.utils import OperationalError, ProgrammingError

from .exceptions import InvalidAssignment, RandomizationListError
from .site_randomizers import site_randomizers

style = color_style()

__all__ = ["RandomizationListVerifier"]


class RandomizationListVerifier:
    """Verifies the Randomization List against the CSV file."""

    def __init__(
        self,
        randomizer_name=None,
        randomizationlist_path: Path | str | None = None,
        randomizer_model_cls=None,
        assignment_map=None,
        fieldnames=None,
        sid_count_for_tests=None,
        required_csv_fieldnames: tuple[str, ...] | None = None,
        **kwargs,  # noqa: ARG002
    ):
        self.count: int = 0
        self.messages: list[str] = []
        self.randomizer_name: str = randomizer_name
        self.randomizer_model_cls = randomizer_model_cls
        self.randomizationlist_path: Path = Path(randomizationlist_path)
        self.assignment_map: dict = assignment_map
        self.sid_count_for_tests: int | None = sid_count_for_tests
        self.required_csv_fieldnames = required_csv_fieldnames

        randomizer_cls = site_randomizers.get(randomizer_name)
        if not randomizer_cls:
            raise RandomizationListError(f"Randomizer not registered. Got `{randomizer_name}`")
        self.fieldnames = fieldnames or self.required_csv_fieldnames
        try:
            self.count = self.randomizer_model_cls.objects.all().count()
        except (ProgrammingError, OperationalError) as e:
            self.messages.append(str(e))
        else:
            if self.count == 0:
                self.messages.append(
                    "Randomization list has not been loaded. "
                    "Run the 'import_randomization_list' management command "
                    "to load before using the system. "
                    "Resolve this issue before using the system."
                )

            elif not self.randomizationlist_path or not self.randomizationlist_path.exists():
                self.messages.append(
                    f"Randomization list file does not exist but SIDs "
                    f"have been loaded. Expected file "
                    f"{self.randomizationlist_path}. "
                    f"Resolve this issue before using the system."
                )
            elif message := self.verify():
                self.messages.append(message)
        if self.messages and (
            "migrate" not in sys.argv
            and "makemigrations" not in sys.argv
            and "import_randomization_list" not in sys.argv
        ):
            raise RandomizationListError(", ".join(self.messages))

    def verify(self) -> str | None:
        message = None
        # read and sort from CSV file
        with self.randomizationlist_path.open(mode="r") as f:
            reader = csv.DictReader(f)
            if self.sid_count_for_tests:
                reader = islice(reader, self.sid_count_for_tests)
            all_rows = [{k: v.strip() for k, v in row.items() if k} for row in reader]
            sorted_rows = sorted(
                all_rows,
                key=lambda row: (row.get("site_name", ""), int(row.get("sid", 0))),
            )
        # compare sorted CSV length with DB
        if len(sorted_rows) != self.randomizer_model_cls.objects.all().count():
            expected_cnt = len(sorted_rows)
            actual_cnt = self.randomizer_model_cls.objects.all().count()
            message = (
                f"Randomization list count is off. Expected {expected_cnt} (CSV). "
                f"Got {actual_cnt} (model_cls). See file "
                f"{self.randomizationlist_path}. "
                f"Resolve this issue before using the system."
            )
        if not message:
            # compare sorted CSV data to DB
            for index, row in enumerate(sorted_rows, start=1):
                sys.stdout.write(f"Index: {index}, SID: {row.get('sid')}, Row: {row}\n")
                message = self.inspect_row(index - 1, row)
                if message:
                    break
        return message

    def inspect_row(self, index: int, row) -> str | None:
        """Checks SIDS, site_name, assignment, ...

        Note:Index is zero-based
        """
        message = None
        obj1 = self.randomizer_model_cls.objects.all().order_by("site_name", "sid")[index]
        try:
            obj2 = self.randomizer_model_cls.objects.get(sid=row["sid"])
        except ObjectDoesNotExist:
            message = f"Randomization file has an invalid SID. Got {row['sid']}"
        else:
            if obj1.sid != obj2.sid:
                message = (
                    f"Randomization list has invalid SIDs. List has invalid SIDs. "
                    f"File data does not match model data. See file "
                    f"{self.randomizationlist_path}. "
                    f"Resolve this issue before using the system. "
                    f"Problem started on line {index + 1}. "
                    f"Got '{row['sid']}' != '{obj1.sid}'."
                )
            if not message:
                assignment = self.get_assignment(row)
                if obj2.assignment != assignment:
                    message = (
                        f"Randomization list does not match model. File data "
                        f"does not match model data. See file "
                        f"{self.randomizationlist_path}. "
                        f"Resolve this issue before using the system. "
                        f"Got '{assignment}' != '{obj2.assignment}' for sid={obj2.sid}."
                    )
                elif obj2.site_name != row["site_name"]:
                    message = (
                        f"Randomization list does not match model. File data "
                        f"does not match model data. See file "
                        f"{self.randomizationlist_path}. "
                        f"Resolve this issue before using the system. "
                        f"Got '{obj2.site_name}' != '{row['site_name']}' "
                        f"for sid={obj2.sid}."
                    )
        return message

    def get_assignment(self, row) -> str:
        """Returns assignment (text) after checking validity."""
        assignment = row["assignment"]
        if assignment not in self.assignment_map:
            raise InvalidAssignment(
                "Invalid assignment. Expected one of "
                f"{list(self.assignment_map.keys())}. "
                f"Got `{assignment}`. "
                f"See randomizer `{self.randomizer_name}`. "
            )
        return assignment

    def get_allocation(self, row) -> int:
        """Returns an integer allocation for the given
        assignment or raises.
        """
        assignment = self.get_assignment(row)
        return self.assignment_map.get(assignment)



from __future__ import annotations

import decimal
import math
from decimal import Decimal


def round_up(value: Decimal, places: Decimal | None = None):
    places = places or Decimal("1.0000")
    if value > Decimal("0.0000000000"):
        decimal.getcontext().rounding = decimal.ROUND_HALF_UP
    else:
        decimal.getcontext().rounding = decimal.ROUND_HALF_DOWN
    return value.quantize(places)


def round_half_up(n, places=0):
    multiplier = 10**places
    return math.floor(n * multiplier + 0.5) / multiplier


def round_half_away_from_zero(n: float | Decimal, places: int | None = None):
    if isinstance(n, Decimal):
        places = Decimal("1" if not places else f"1.{str(0) * places}")
        return round_up(n, places=places)
    rounded_abs = round_half_up(abs(n), places)
    return math.copysign(rounded_abs, n)



from django.db import models

from edc_constants.choices import YES_NO, YES_NO_DECLINED

from ..validators import eligible_if_yes, eligible_if_yes_or_declined


class ReviewFieldsMixin(models.Model):
    consent_reviewed = models.CharField(
        verbose_name="I have reviewed the consent with the participant",
        max_length=3,
        choices=YES_NO,
        validators=[eligible_if_yes],
        null=True,
        blank=False,
        help_text="If no, participant is not eligible.",
    )

    study_questions = models.CharField(
        verbose_name="I have answered all questions the participant had about the study",
        max_length=3,
        choices=YES_NO,
        validators=[eligible_if_yes],
        null=True,
        blank=False,
        help_text="If no, participant is not eligible.",
    )

    assessment_score = models.CharField(
        verbose_name=(
            "I have asked the participant questions about this study and "
            "the participant has demonstrated understanding"
        ),
        max_length=3,
        choices=YES_NO,
        validators=[eligible_if_yes],
        null=True,
        blank=False,
        help_text="If no, participant is not eligible.",
    )

    consent_signature = models.CharField(
        verbose_name="I have verified that the participant has signed the consent form",
        max_length=3,
        choices=YES_NO,
        validators=[eligible_if_yes],
        null=True,
        blank=False,
        help_text="If no, participant is not eligible.",
    )

    consent_copy = models.CharField(
        verbose_name=(
            "I have provided the participant with a copy of their signed informed consent"
        ),
        max_length=20,
        choices=YES_NO_DECLINED,
        validators=[eligible_if_yes_or_declined],
        null=True,
        blank=False,
        help_text="If declined, return copy with the consent",
    )

    class Meta:
        abstract = True



from django.contrib import admin
from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin

from .admin_site import edc_refusal_admin
from .forms import SubjectRefusalForm
from .models import SubjectRefusal


class SubjectRefusalModelAdminMixin:
    fieldsets = (
        [
            None,
            {
                "fields": (
                    "screening_identifier",
                    "report_datetime",
                    "reason",
                    "other_reason",
                    "comment",
                )
            },
        ],
        audit_fieldset_tuple,
    )

    list_display = (
        "screening_identifier",
        "report_datetime",
        "reason",
        "user_created",
        "created",
    )

    list_filter = ("report_datetime", "reason")

    search_fields = ("screening_identifier",)

    radio_fields = {"reason": admin.VERTICAL}


@admin.register(SubjectRefusal, site=edc_refusal_admin)
class SubjectRefusalAdmin(
    SiteModelAdminMixin,
    SubjectRefusalModelAdminMixin,
    ModelAdminSubjectDashboardMixin,
    SimpleHistoryAdmin,
):
    form = SubjectRefusalForm

    post_url_on_delete_name = "screening_listboard_url"
    subject_listboard_url_name = "screening_listboard_url"
    subject_dashboard_url_name = "screening_listboard_url"

    def get_subject_dashboard_url_kwargs(self, obj):
        return dict(screening_identifier=obj.screening_identifier)

    def view_on_site(self, obj):
        try:
            url = reverse(
                self.get_subject_dashboard_url_name(),
                kwargs=self.get_subject_dashboard_url_kwargs(obj),
            )
        except NoReverseMatch as e:
            if callable(super().view_on_site):
                url = super().view_on_site(obj)
            else:
                raise NoReverseMatch(f"{e}. See subject_dashboard_url_name for {self!r}.")
        return url



from typing import Any

from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.utils.html import escape
from django.views.generic.base import ContextMixin

from edc_lab.models import Box, Manifest, ManifestItem


class ManifestItemError(Exception):
    pass


class ManifestViewMixin(ContextMixin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._manifest = None
        self._manifest_item = None
        self._manifest_identifier = None
        self._manifest_item_identifier = None
        self.original_manifest_item_identifier = None
        self.original_manifest_identifier = None

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            {
                "manifest_identifier": self.original_manifest_identifier,
                "manifest_item_identifier": self.original_manifest_item_identifier,
                "manifest": self.manifest,
            }
        )
        return super().get_context_data(**kwargs)

    @property
    def manifest_identifier(self):
        if not self._manifest_identifier:
            self.original_manifest_identifier = escape(
                self.kwargs.get("manifest_identifier")
            ).strip()
            self._manifest_identifier = "".join(self.original_manifest_identifier.split("-"))
        return self._manifest_identifier

    @property
    def manifest_item_identifier(self):
        """Returns a cleaned manifest_item_identifier or None."""
        if not self._manifest_item_identifier:
            self.original_manifest_item_identifier = escape(
                self.request.POST.get("manifest_item_identifier", "")
            ).strip()
            if self.original_manifest_item_identifier:
                self._manifest_item_identifier = self._clean_manifest_item_identifier()
        return self._manifest_item_identifier

    @property
    def manifest(self):
        if not self._manifest:
            if self.manifest_identifier:
                try:
                    self._manifest = Manifest.objects.get(
                        manifest_identifier=self.manifest_identifier
                    )
                except ObjectDoesNotExist:
                    self._manifest = None
        return self._manifest

    @property
    def manifest_item(self):
        """Returns a manifest item model instance."""
        if not self._manifest_item:
            if self.manifest_item_identifier:
                try:
                    self._manifest_item = ManifestItem.objects.get(
                        manifest=self.manifest, identifier=self.manifest_item_identifier
                    )
                except ObjectDoesNotExist:
                    messages.error(
                        self.request,
                        f"Invalid manifest item. Got {self.original_manifest_item_identifier}",
                    )
        return self._manifest_item

    def get_manifest_item(self, position):
        """Returns a manifest item model instance for the given position."""
        try:
            manifest_item = ManifestItem.objects.get(manifest=self.manifest, position=position)
        except ObjectDoesNotExist:
            messages.error(self.request, f"Invalid position for manifest. Got {position}")
            return None
        return manifest_item

    def _clean_manifest_item_identifier(self):
        """Returns a valid identifier or raises."""
        manifest_item_identifier = "".join(self.original_manifest_item_identifier.split("-"))
        try:
            Box.objects.get(box_identifier=manifest_item_identifier)
        except ObjectDoesNotExist:
            messages.error(
                self.request,
                f'Invalid box. Got {self.original_manifest_item_identifier or "None"}.',
            )
        return manifest_item_identifier



from typing import Any

from django.apps import apps as django_apps

from .get_default_navbar import get_default_navbar
from .site_navbars import site_navbars


class NavbarViewMixin:
    navbar_selected_item = None
    navbar_name = get_default_navbar()

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        """Add rendered navbar <navbar_name> to the context for
        this view.

        Also adds the "default" navbar.
        """
        kwargs = self.get_navbar_context_data(kwargs)
        return super().get_context_data(**kwargs)

    def get_navbar_name(self):
        return self.navbar_name

    def get_navbar_context_data(self, context) -> dict:
        navbar = site_navbars.get_navbar(name=self.get_navbar_name())
        navbar.set_active(self.get_navbar_selected(**context))
        context.update(navbar=navbar)
        app_config = django_apps.get_app_config("edc_navbar")
        default_navbar_name = app_config.default_navbar_name
        if default_navbar_name and self.get_navbar_name() != default_navbar_name:
            default_navbar = site_navbars.get_navbar(name=default_navbar_name)
            default_navbar.set_active(self.navbar_selected_item)
            context.update(
                default_navbar=default_navbar, default_navbar_name=default_navbar_name
            )
        return context

    def get_navbar_selected(self, **kwargs) -> str:
        return self.navbar_selected_item



import pandas as pd

from edc_constants.constants import FEMALE, MALE

from ..constants import N_ONLY, N_WITH_COL_PROP, N_WITH_ROW_PROP
from ..row import RowDefinition, RowDefinitions
from ..table import Table


class FbgOgttTable1(Table):

    fbg_colname = "fbg"
    ogtt_colname = "ogtt"

    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(
            colname="",
            main_df=main_df,
            title="OGTT & FBG (mmol/L) categories",
        )

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=False)
        row0 = RowDefinition(
            title=self.title,
            label=self.default_sublabel,
            condition=(df_tmp["gender"].notna()),
            columns={FEMALE: (N_ONLY, 2), MALE: (N_ONLY, 2), "All": (N_ONLY, 2)},
            drop=False,
        )
        row_defs.add(row0)
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }

        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="Not fasted",
                condition=(
                    (self.main_df["fasting_fbg_hrs"] < 8.0)
                    | (self.main_df["fasting_ogtt_hrs"] < 8.0)
                ),
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.ogtt_colname] >= 11.1)
            & (self.main_df[self.fbg_colname] >= 7.0)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT ≥11.1 and FBG ≥7.0",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            ((df_tmp[self.ogtt_colname] >= 11.1) | (df_tmp[self.fbg_colname] >= 7.0))
            & (df_tmp[self.fbg_colname].notna())
            & (df_tmp[self.ogtt_colname].notna())
        )

        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT ≥11.1 or FBG ≥7.0",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            ((df_tmp[self.fbg_colname] >= 6.1) | (df_tmp[self.ogtt_colname] >= 7.8))
            & (df_tmp[self.fbg_colname].notna())
            & (df_tmp[self.ogtt_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT ≥7.8 or FBG ≥6.1",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            ((df_tmp[self.fbg_colname] < 6.1) | (df_tmp[self.ogtt_colname] < 7.8))
            & (df_tmp[self.fbg_colname].notna())
            & (df_tmp[self.ogtt_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT <7.8 or FBG <6.1",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (df_tmp[self.fbg_colname].notna()) & (df_tmp[self.ogtt_colname].notna())
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="Other",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (df_tmp[self.fbg_colname].notna()) & (df_tmp[self.ogtt_colname].isna())
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT not measured",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )
        condition = (df_tmp[self.fbg_colname].isna()) & (df_tmp[self.ogtt_colname].notna())
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG not measured",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )
        condition = (df_tmp[self.fbg_colname].isna()) & (df_tmp[self.ogtt_colname].isna())
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="not measured",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )
        return row_defs


class FbgOgttTable2(Table):

    fbg_colname = "fbg"
    ogtt_colname = "ogtt"

    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(
            colname="",
            main_df=main_df,
            title="OGTT & FBG (mmol/L) additional",
        )

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=False)
        row0 = RowDefinition(
            title=self.title,
            label="Glucose levels, n (%)",
            condition=(df_tmp["gender"].notna()),
            columns={FEMALE: (N_ONLY, 2), MALE: (N_ONLY, 2), "All": (N_ONLY, 2)},
            drop=False,
        )
        row_defs.add(row0)
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }

        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="Not fasted",
                condition=(
                    (self.main_df["fasting_fbg_hrs"] < 8.0)
                    | (self.main_df["fasting_ogtt_hrs"] < 8.0)
                ),
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] < 6.1)
            & (self.main_df[self.ogtt_colname] < 7.8)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG <6.1 mmol/l and after OGTT <7.8 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] < 6.1)
            & (self.main_df[self.ogtt_colname] >= 7.8)
            & (self.main_df[self.ogtt_colname] < 11.1)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG <6.1 mmol/l and after OGTT 7.8–11.0 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] >= 6.1)
            & (self.main_df[self.fbg_colname] < 7.0)
            & (self.main_df[self.ogtt_colname] < 7.8)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG 6.1–6.9 mmol/l and after OGTT <7.8 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] >= 6.1)
            & (self.main_df[self.fbg_colname] < 7.0)
            & (self.main_df[self.ogtt_colname] >= 7.8)
            & (self.main_df[self.ogtt_colname] < 11.1)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG 6.1–6.9 mmol/l and after OGTT 7.8–11.0 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] >= 6.1)
            & (self.main_df[self.fbg_colname] < 7.0)
            & (self.main_df[self.ogtt_colname] >= 11.1)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG 6.1–6.9 mmol/l and after OGTT ≥11.0 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] >= 7.0)
            & (self.main_df[self.ogtt_colname] < 7.8)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG ≥7.0 mmol/l and after OGTT <7.8 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] >= 7.0)
            & (self.main_df[self.ogtt_colname] >= 7.8)
            & (self.main_df[self.ogtt_colname] < 11.1)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG ≥7.0 mmol/l and after OGTT 7.8–11.0 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (
            (self.main_df[self.fbg_colname] >= 7.0)
            & (self.main_df[self.ogtt_colname] >= 11.1)
            & (self.main_df[self.ogtt_colname].notna())
            & (self.main_df[self.fbg_colname].notna())
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG ≥7.0 mmol/l and after OGTT ≥11.1 mmol/l",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (self.main_df[self.ogtt_colname].notna()) | (
            self.main_df[self.fbg_colname].isna()
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="FBG not measured",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (self.main_df[self.ogtt_colname].isna()) | (
            self.main_df[self.fbg_colname].notna()
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="OGTT not measured",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )

        condition = (self.main_df[self.ogtt_colname].isna()) | (
            self.main_df[self.fbg_colname].isna()
        )
        row_defs.add(
            RowDefinition(
                colname=self.colname,
                label="not measured",
                condition=condition,
                columns=columns,
                drop=True,
            )
        )
        return row_defs


class FbgOgttTable(Table):

    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(colname="", main_df=main_df, title="")

    def build_table_df(self) -> None:
        df1 = FbgOgttTable1(main_df=self.main_df).table_df
        df2 = FbgOgttTable2(main_df=self.main_df).table_df
        self.table_df = pd.concat([df1, df2])



from django.contrib import admin

from edc_list_data.admin import ListModelAdminMixin

from .admin_site import edc_adverse_event_admin
from .models import AeClassification, CauseOfDeath, SaeReason


@admin.register(AeClassification, site=edc_adverse_event_admin)
class AeClassificationAdmin(ListModelAdminMixin, admin.ModelAdmin):
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"


@admin.register(CauseOfDeath, site=edc_adverse_event_admin)
class CauseOfDeathAdmin(ListModelAdminMixin, admin.ModelAdmin):
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"


@admin.register(SaeReason, site=edc_adverse_event_admin)
class SaeReasonAdmin(ListModelAdminMixin, admin.ModelAdmin):
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"



from .data_request import DataRequest
from .data_request_history import DataRequestHistory
from .edc_permissions import EdcPermissions
from .export_receipt import ExportReceipt
from .file_history import FileHistory
from .object_history import ObjectHistory
from .permission_dummies import ExportData, ImportData
from .plan import Plan
from .signals import (
    export_transaction_history_on_post_save,
    export_transaction_history_on_pre_delete,
)
from .upload_export_receipt_file import UploadExportReceiptFile



# Generated by Django 5.2.3 on 2025-07-04 13:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0078_alter_historicalstock_qty_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalstock",
            name="allocated",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="allocated",
        ),
    ]



from edc_action_item.action_with_notification import ActionWithNotification
from edc_constants.constants import HIGH_PRIORITY
from edc_offstudy.constants import END_OF_STUDY_ACTION
from edc_visit_tracking.constants import VISIT_MISSED_ACTION

from .constants import LTFU_ACTION
from .utils import get_ltfu_model_name


class LtfuAction(ActionWithNotification):
    reference_model = get_ltfu_model_name()
    admin_site_name = "edc_ltfu_admin"

    name = LTFU_ACTION
    display_name = "Submit Loss to Follow Up Report"
    notification_display_name = "Loss to Follow Up Report"
    parent_action_names = (VISIT_MISSED_ACTION,)
    show_link_to_changelist = True
    priority = HIGH_PRIORITY

    def get_next_actions(self):
        next_actions = [END_OF_STUDY_ACTION]
        return next_actions



from __future__ import annotations

from typing import TYPE_CHECKING

from .subquery import Subquery

if TYPE_CHECKING:
    from .qa_case import QaCase


def subquery_from_dict(
    cases: list[dict[str:str, str:str, str:str] | QaCase],
    as_list: bool | None = False,
) -> str | list:
    """Returns an SQL select statement as a union of the select
    statements of each case.

    args:
     cases = [{
         "label_lower": "my_app.hivhistory",
         "dbtable": "my_app_hivhistory",
         "field": "hiv_init_date",
         "label": "missing HIV initiation date",
         "list_tables": [(list_field, list_dbtable, alias), ...],
         }, ...]

         Note: `list_field` is the CRF id field, for example:
            left join <list_dbtable> as <alias> on crf.<list_field>=<alias>.id
    """
    subqueries = []
    for case in cases:
        try:
            subquery = case.sql
        except AttributeError:
            subquery = Subquery(**case).sql
        subqueries.append(subquery)
    if as_list:
        return subqueries
    return " UNION ".join(subqueries)



# Generated by Django 3.2.6 on 2021-09-21 01:03

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0019_auto_20210510_2036"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataquery",
            name="locked_reason",
            field=models.TextField(
                blank=True,
                help_text="If required, the reason the query cannot be resolved.",
                null=True,
                verbose_name="Reason query locked",
            ),
        ),
        migrations.AddField(
            model_name="historicaldataquery",
            name="locked_reason",
            field=models.TextField(
                blank=True,
                help_text="If required, the reason the query cannot be resolved.",
                null=True,
                verbose_name="Reason query locked",
            ),
        ),
    ]



from django.apps import apps as django_apps
from django.contrib import admin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import ModelAdminInstitutionMixin, TemplatesModelAdminMixin
from edc_sites.admin import SiteModelAdminMixin

from ..admin_site import edc_qareports_admin
from ..models import QaReportLog


@admin.register(QaReportLog, site=edc_qareports_admin)
class QaReportLogAdmin(
    SiteModelAdminMixin,
    ModelAdminRevisionMixin,
    ModelAdminInstitutionMixin,
    TemplatesModelAdminMixin,
    admin.ModelAdmin,
):
    ordering = ["-accessed"]

    list_display = ["report", "username", "site", "accessed", "report_model"]

    list_filter = ["accessed", "report_model", "username", "site"]

    search_fields = ["report_model", "username"]

    readonly_fields = ["report_model", "username", "site", "accessed"]

    @admin.display(description="Report", ordering="report_model")
    def report(self, obj=None):
        model_cls = django_apps.get_model(obj.report_model)
        return model_cls._meta.verbose_name



from __future__ import annotations

import getpass
import os.path
import sys

from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.core.management import CommandError, color_style
from django.core.management.base import BaseCommand

from edc_model_to_dataframe.model_to_dataframe import ModelToDataframe
from edc_pdutils.df_exporters import Exporter
from edc_pdutils.utils import get_model_names
from edc_sites.site import sites

ALL_COUNTRIES = "all"

style = color_style()


class Command(BaseCommand):
    def __init__(self, **kwargs):
        self.decrypt: bool | None = None
        self.site_ids: list[int] = []
        self.exclude_historical: bool | None = None
        super().__init__(**kwargs)

    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--app",
            dest="app_labels",
            default="",
            help="app label. Separate by comma if more than one.",
        )

        parser.add_argument(
            "-m",
            "--model",
            dest="model_names",
            default="",
            help="model name in label_lower format. Separate by comma if more than one.",
        )

        parser.add_argument(
            "--skip_model",
            dest="skip_model_names",
            default="",
            help="models to skip in label_lower format. Separate by comma if more than one.",
        )

        parser.add_argument(
            "-p",
            "--path",
            dest="path",
            default=False,
            help="export path",
        )

        parser.add_argument(
            "-f",
            "--format",
            dest="format",
            default="csv",
            choices=["csv", "stata"],
            help="export format (csv, stata)",
        )

        parser.add_argument(
            "--stata-dta-version",
            dest="stata_dta_version",
            default=None,
            choices=["118", "119"],
            help="STATA DTA file format version",
        )

        parser.add_argument(
            "--include-historical",
            action="store_true",
            dest="include_historical",
            default=False,
            help="export historical tables",
        )

        parser.add_argument(
            "--decrypt",
            action="store_true",
            dest="decrypt",
            default=False,
            help="decrypt",
        )

        parser.add_argument(
            "--use-simple-filename",
            action="store_true",
            dest="use_simple_filename",
            default=False,
            help="do not use app_label or datestamp in filename",
        )

        parser.add_argument(
            "--country",
            dest="countries",
            default="",
            help=(
                "only export data for country. Separate by comma if more than one. "
                f"Use `{ALL_COUNTRIES}` to export all countries."
            ),
        )

        parser.add_argument(
            "--site",
            dest="site_ids",
            default="",
            help="only export data for site id. Separate by comma if more than one.",
        )

    def handle(self, *args, **options):
        self.validate_user_perms_or_raise()

        date_format = "%Y-%m-%d %H:%M:%S"
        sep = "|"
        export_format = options["format"]
        export_path = options["path"]
        stata_dta_version = options["stata_dta_version"]
        if not export_path or not os.path.exists(export_path):
            raise CommandError(f"Path does not exist. Got `{export_path}`")
        use_simple_filename = options["use_simple_filename"]
        self.exclude_historical = not options["include_historical"]
        self.decrypt = options["decrypt"]

        # TODO: inspect username that you are preparing data for
        site_ids = options["site_ids"] or []
        if site_ids:
            site_ids = options["site_ids"].split(",")
        countries = self.get_countries(options)
        self.site_ids = self.get_site_ids(site_ids=site_ids, countries=countries)
        app_labels = options["app_labels"] or []
        if app_labels:
            app_labels = options["app_labels"].split(",")
        model_names = options["model_names"] or []
        if model_names:
            model_names = options["model_names"].split(",")
        skip_model_names = []
        if options["skip_model_names"]:
            skip_model_names = options["skip_model_names"].split(",")
        if app_labels and model_names:
            raise CommandError(
                "Either provide the `app label` or a `model name` but not both. "
                f"Got {app_labels} and {model_names}."
            )
        models = self.get_models(app_labels=app_labels, model_names=model_names)
        if not models:
            raise CommandError("Nothing to do. No models to export.")
        self.export(
            models,
            skip_model_names,
            date_format,
            sep,
            export_path,
            use_simple_filename,
            export_format,
            stata_dta_version,
        )

    def export(
        self,
        models,
        skip_model_names,
        date_format,
        sep,
        export_path,
        use_simple_filename,
        export_format,
        stata_dta_version,
    ):
        for app_label, model_names in models.items():
            for model_name in model_names:
                if model_name in skip_model_names:
                    continue
                try:
                    m = ModelToDataframe(
                        model=model_name,
                        drop_sys_columns=False,
                        decrypt=self.decrypt,
                        sites=self.site_ids,
                    )
                except LookupError as e:
                    sys.stdout.write(style.ERROR(f"     LookupError: {e}\n"))
                else:
                    exporter = Exporter(
                        model_name=model_name,
                        date_format=date_format,
                        delimiter=sep,
                        export_folder=export_path,
                        app_label=app_label,
                        use_simple_filename=use_simple_filename,
                    )
                    if not export_format or export_format == "csv":
                        exporter.to_csv(dataframe=m.dataframe)
                    elif export_format == "stata":
                        exporter.to_stata(dataframe=m.dataframe, dta_version=stata_dta_version)
                    print(f" * {model_name}")

    def validate_user_perms_or_raise(self) -> None:
        username = input("Username:")
        passwd = getpass.getpass("Password for " + username + ":")
        try:
            user = User.objects.get(username=username, is_superuser=False, is_active=True)
        except ObjectDoesNotExist:
            raise CommandError("Invalid username or password.")
        if not user.check_password(passwd):
            raise CommandError("Invalid username or password.")
        if not user.groups.filter(name="EXPORT").exists():
            raise CommandError("You are not authorized to export data.")
        if self.decrypt and not user.groups.filter(name="EXPORT_PII").exists():
            raise CommandError("You are not authorized to export sensitive data.")

    @staticmethod
    def get_countries(options):
        countries = options["countries"] or []
        if not countries:
            raise CommandError("Expected country.")
        if countries == ALL_COUNTRIES:
            countries = sites.countries
        else:
            countries = options["countries"].lower().split(",")
            for country in countries:
                if country not in sites.countries:
                    raise CommandError(f"Invalid country. Got {country}.")
        return countries

    def get_models(
        self, app_labels: list[str] | None, model_names: list[str] | None
    ) -> dict[str, list[str]]:
        models = {}
        if model_names:
            for model_name in model_names:
                app_label, _ = model_name.split(".")
                if self.exclude_historical and model_name.startswith("historical"):
                    continue
                try:
                    models[app_label].append(model_name)
                except KeyError:
                    models[app_label] = [model_name]
        elif app_labels:
            for app_label in app_labels:
                models.update(
                    {
                        app_label: get_model_names(
                            app_label=app_label,
                            exclude_historical=self.exclude_historical,
                        )
                    }
                )
        return models

    @staticmethod
    def get_site_ids(
        site_ids: list[int] | list[str] | None,
        countries: list[str] | None,
    ) -> list[int]:
        if countries and site_ids:
            raise CommandError("Invalid. Specify `site_ids` or `countries`, not both.")
        for site_id in site_ids or []:
            try:
                obj = Site.objects.get(id=int(site_id))
            except ObjectDoesNotExist:
                raise CommandError(f"Invalid site_id. Got `{site_id}`.")
            else:
                site_ids.append(obj.id)
        for country in countries or []:
            site_ids.extend(list(sites.get_by_country(country)))
        return site_ids



from __future__ import annotations

from django.db import models

from edc_model.models import BaseUuidModel, HistoricalRecords

from .reference_model_mixins import ReferenceModelMixin


class GradingData(ReferenceModelMixin, BaseUuidModel):

    grade = models.IntegerField()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.description} GRADE {self.grade}"

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Grading Reference"
        verbose_name_plural = "Grading References"



from __future__ import annotations

from django.conf import settings
from django_audit_fields.admin import ModelAdminAuditFieldsMixin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminReplaceLabelTextMixin,
    TemplatesModelAdminMixin,
)
from edc_notification.modeladmin_mixins import NotificationModelAdminMixin

from .model_admin_dashboard_mixin import ModelAdminDashboardMixin


class ModelAdminSubjectDashboardMixin(
    ModelAdminDashboardMixin,
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,  # add
    NotificationModelAdminMixin,
    ModelAdminFormInstructionsMixin,  # add
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminReplaceLabelTextMixin,
    ModelAdminAuditFieldsMixin,
):
    date_hierarchy = "modified"
    empty_value_display = "-"
    list_per_page = 10
    show_cancel = True

    class Media:
        js = ("edc_model_admin/admin/js/delay_save_buttons.js",)

    def get_list_filter(self, request) -> tuple[str, ...]:
        return super().get_list_filter(request)

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        return super().get_readonly_fields(request, obj=obj)

    def get_search_fields(self, request) -> tuple[str, ...]:
        return super().get_search_fields(request)

    def changeform_view(self, request, object_id=None, form_url="", extra_context=None):
        extra_context = extra_context or {}
        extra_context["save_delay"] = getattr(
            settings, "EDC_MODEL_ADMIN_SAVE_DELAY", 0
        )  # milliseconds
        return super().changeform_view(
            request, object_id, form_url, extra_context=extra_context
        )



from warnings import warn

from django.db import models

from ...validators import dh_validator, ymd_validator


class DurationDHField(models.CharField):
    description = "Duration in d/h"

    def __init__(self, *args, **kwargs) -> None:
        kwargs["verbose_name"] = kwargs.get("verbose_name") or "Duration:"
        kwargs["max_length"] = 7
        kwargs["validators"] = [dh_validator]
        kwargs["help_text"] = (
            f"{kwargs.get('help_text') or ''} Format is `DDdHHh`, `DDd` or `HHh`. "
            "For example 1d23h, 15d9h ... or 20d, or 5h ..."
        )
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["verbose_name"]
        del kwargs["max_length"]
        del kwargs["validators"]
        del kwargs["help_text"]
        return name, path, args, kwargs


class DurationYMDField(models.CharField):
    description = "Duration in y/m"

    def __init__(self, *args, **kwargs) -> None:
        kwargs["verbose_name"] = kwargs.get("verbose_name") or "Duration:"
        kwargs["max_length"] = 8
        kwargs["validators"] = [ymd_validator]
        kwargs["help_text"] = (
            f"{kwargs.get('help_text') or ''} Format is `YYyMMm` or `DDd`. "
            "For example 3y10m, 12y7m ... or 7d, 0d ..."
        )
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["verbose_name"]
        del kwargs["max_length"]
        del kwargs["validators"]
        del kwargs["help_text"]
        return name, path, args, kwargs


class DurationYearMonthField(DurationYMDField):
    def __init__(self, *args, **kwargs) -> None:
        warn(
            "DurationYearMonthField has been deprecated. Use DurationYMDField instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



from django.forms import ValidationError

from edc_constants.constants import NOT_APPLICABLE, OTHER

from .base_form_validator import (
    APPLICABLE_ERROR,
    NOT_APPLICABLE_ERROR,
    NOT_REQUIRED_ERROR,
    REQUIRED_ERROR,
    BaseFormValidator,
)

M2M_SELECTION_ONLY = "m2m_selection_only"
M2M_INVALID_SELECTION = "m2m_invalid_selection"
M2M_INVALID_COMBINATION = "m2m_invalid_combination"


class ManyToManyFieldValidator(BaseFormValidator):
    def get_m2m_selected(self, m2m_field: str) -> dict:
        qs = self.cleaned_data.get(m2m_field) or []
        return {
            getattr(obj, self.default_fk_stored_field_name): getattr(
                obj, self.default_fk_display_field_name
            )
            for obj in qs
        }

    def m2m_applicable_if_true(self, cond: bool, m2m_field: str = None) -> bool:
        code: str | None = None
        message: dict = {}
        qs = self.cleaned_data.get(m2m_field)
        if cond:
            if qs and qs.count() > 0:
                selected = self.get_m2m_selected(m2m_field)
                if NOT_APPLICABLE in selected:
                    message = {m2m_field: "This field is applicable"}
                    code = APPLICABLE_ERROR
        elif qs and qs.count() > 0:
            selected = self.get_m2m_selected(m2m_field)
            if NOT_APPLICABLE not in selected:
                message = {m2m_field: "This field is not applicable"}
                code = NOT_APPLICABLE_ERROR
        if message:
            self._errors.update(message)
            self._error_codes.append(code)
            raise ValidationError(message, code=code)
        return False

    def m2m_applicable_if(
        self,
        *responses: str | int | bool,
        field: str = None,
        m2m_field: str = None,
    ) -> bool:
        """Raises an exception or returns False.

        m2m_field is applicable if field is in responses.
        """
        code: str | None = None
        message: dict = {}
        if self.cleaned_data.get(field):
            qs = self.cleaned_data.get(m2m_field)
            if self.cleaned_data.get(field) not in responses:
                # m2m should == NOT_APPLICABLE
                if qs and qs.count() > 0:
                    selected = self.get_m2m_selected(m2m_field)
                    if NOT_APPLICABLE not in selected:
                        message = {m2m_field: "This field is not applicable"}
                        code = NOT_APPLICABLE_ERROR
            # m2m should != NOT_APPLICABLE
            elif qs and qs.count() > 0:
                selected = self.get_m2m_selected(m2m_field)
                if NOT_APPLICABLE in selected:
                    message = {m2m_field: "This field is applicable"}
                    code = APPLICABLE_ERROR
        if message:
            self._errors.update(message)
            self._error_codes.append(code)
            raise ValidationError(message, code=code)
        return False

    def m2m_required(self, m2m_field: str = None) -> bool:
        """Raises an exception or returns False.

        m2m_field is required.
        """
        code: str | None = None
        message: dict = {}
        if not self.cleaned_data.get(m2m_field):
            message = {m2m_field: "This field is required"}
            code = REQUIRED_ERROR
        if message:
            self._errors.update(message)
            self._error_codes.append(code)
            raise ValidationError(message, code=code)
        return False

    def m2m_not_required(self, m2m_field: str = None) -> bool:
        """Raises an exception or returns False.

        m2m_field is not required.
        """
        code: str | None = None
        message: dict = {}
        if self.cleaned_data.get(m2m_field):
            message = {m2m_field: "This field is not required"}
            code = REQUIRED_ERROR
        if message:
            self._errors.update(message)
            self._error_codes.append(code)
            raise ValidationError(message, code=code)
        return False

    def m2m_required_if(
        self, response: str | int | bool, field: str = None, m2m_field: str = None
    ) -> bool:
        """Raises an exception or returns False.

        m2m_field is required if field  == response
        """
        code: str | None = None
        message: dict = {}
        if (
            self.cleaned_data.get(field) == response and not self.cleaned_data.get(m2m_field)
        ) or (
            self.cleaned_data.get(field) == response
            and self.cleaned_data.get(m2m_field).count() == 0
        ):
            message = {m2m_field: "This field is required"}
            code = REQUIRED_ERROR
        elif (
            self.cleaned_data.get(field) != response
            and self.cleaned_data.get(m2m_field)
            and self.cleaned_data.get(m2m_field).count() != 0
        ):
            message = {m2m_field: "This field is not required"}
            code = NOT_REQUIRED_ERROR
        if message:
            self._errors.update(message)
            self._error_codes.append(code)
            raise ValidationError(message, code=code)
        return False

    def m2m_single_selection_if(self, *single_selections: str, m2m_field=None):
        """Raises an exception or returns False.

        if a selected response from m2m_field is in single_selections
        and there is more than one selected value, raises.
        """
        qs = self.cleaned_data.get(m2m_field)
        if qs and qs.count() > 1:
            selected = self.get_m2m_selected(m2m_field)
            for selection in single_selections:
                if selection in selected:
                    message = {
                        m2m_field: f"Invalid combination. '{selected.get(selection)}' "
                        f"may not be combined "
                        f"with other selections"
                    }
                    self._errors.update(message)
                    self._error_codes.append(M2M_INVALID_COMBINATION)
                    raise ValidationError(message, code=M2M_INVALID_COMBINATION)
        return False

    def required_if_m2m(
        self,
        *responses: str | int | bool,
        field: str = None,
        field_required: str | None = None,
        field_other_evaluate_as_int: bool | None = None,
    ) -> bool:
        """Same as required_if() but where field is an m2m field."""
        return self.m2m_other_specify(
            *responses,
            m2m_field=field,
            field_other=field_required,
            field_other_evaluate_as_int=field_other_evaluate_as_int,
        )

    def m2m_other_specify(
        self,
        *responses: str | int | bool,
        m2m_field: str = None,
        field_other: str | None = None,
        field_other_evaluate_as_int: bool | None = None,
    ) -> bool:
        """Raises an exception or returns False.

        field_other is required if a selected response from m2m_field
        is in responses

        Note: for edc list models, "name" is the stored value!
        """
        qs = self.cleaned_data.get(m2m_field)
        found = False
        responses = (OTHER,) if len(responses) == 0 else responses

        if field_other_evaluate_as_int:
            field_other_has_value = self.cleaned_data.get(field_other) is not None
        else:
            field_other_has_value = self.cleaned_data.get(field_other)
        if qs and qs.count() > 0:
            for response in responses:
                if response in self.get_m2m_selected(m2m_field):
                    found = True
            if found and not field_other_has_value:
                message = {field_other: "This field is required."}
                self._errors.update(message)
                self._error_codes.append(REQUIRED_ERROR)
                raise ValidationError(message, code=REQUIRED_ERROR)
            if not found and field_other_has_value:
                message = {field_other: "This field is not required."}
                self._errors.update(message)
                self._error_codes.append(NOT_REQUIRED_ERROR)
                raise ValidationError(message, code=NOT_REQUIRED_ERROR)
        elif field_other_has_value:
            message = {field_other: "This field is not required."}
            self._errors.update(message)
            self._error_codes.append(NOT_REQUIRED_ERROR)
            raise ValidationError(message, code=NOT_REQUIRED_ERROR)
        return False

    def m2m_other_not_specify(
        self,
        *responses: str | int | bool,
        m2m_field: str = None,
        field_other: str | None = None,
    ) -> bool:
        """Raises an exception or returns False.

        field_other is NOT required if a selected response from m2m_field
        is in responses

        Note: for edc list models, "name" is the stored value!
        """
        qs = self.cleaned_data.get(m2m_field)
        found = False
        if qs and qs.count() > 0:
            for response in responses:
                if response in self.get_m2m_selected(m2m_field):
                    found = True
            if found and self.cleaned_data.get(field_other):
                message = {field_other: "This field is not required."}
                self._errors.update(message)
                self._error_codes.append(NOT_REQUIRED_ERROR)
                raise ValidationError(message, code=NOT_REQUIRED_ERROR)
            if not found and not self.cleaned_data.get(field_other):
                message = {field_other: "This field is required."}
                self._errors.update(message)
                self._error_codes.append(REQUIRED_ERROR)
                raise ValidationError(message, code=REQUIRED_ERROR)
        elif self.cleaned_data.get(field_other):
            message = {field_other: "This field is not required."}
            self._errors.update(message)
            self._error_codes.append(NOT_REQUIRED_ERROR)
            raise ValidationError(message, code=NOT_REQUIRED_ERROR)
        return False

    def m2m_other_specify_applicable(
        self,
        *responses: str | int | bool,
        m2m_field: str = None,
        field_other: str | None = None,
    ) -> bool:
        """Raises an exception or returns False.

        field_other is applicable if a selected response from m2m_field
        is in responses
        """
        selected = self.cleaned_data.get(m2m_field)
        found = False
        if selected and selected.count() > 0:
            for response in responses:
                if response in [
                    getattr(o, self.default_fk_stored_field_name) for o in selected
                ]:
                    found = True
            if found and self.cleaned_data.get(field_other) == NOT_APPLICABLE:
                message = {field_other: "This field is applicable."}
                self._errors.update(message)
                self._error_codes.append(APPLICABLE_ERROR)
                raise ValidationError(message, code=APPLICABLE_ERROR)
            if not found and self.cleaned_data.get(field_other) != NOT_APPLICABLE:
                message = {field_other: "This field is not applicable."}
                self._errors.update(message)
                self._error_codes.append(NOT_APPLICABLE_ERROR)
                raise ValidationError(message, code=NOT_APPLICABLE_ERROR)
        elif self.cleaned_data.get(field_other) != NOT_APPLICABLE:
            message = {field_other: "This field is not applicable."}
            self._errors.update(message)
            self._error_codes.append(NOT_APPLICABLE_ERROR)
            raise ValidationError(message, code=NOT_APPLICABLE_ERROR)
        return False

    def m2m_selection_expected(
        self,
        response: str | int | bool,
        m2m_field: str = None,
        error_msg: str = None,
    ) -> bool:
        """Raises an exception or returns False.

        m2m_field is required and the selection must be `response`
        and only `response`.

        This would normally be preceded by an IF condition in the code.

            if (condition):
                self.m2m_selection_expected(...)
        """
        qs = self.cleaned_data.get(m2m_field)
        if qs and qs.count() > 0:
            selection = [
                getattr(obj, self.default_fk_stored_field_name)
                for obj in qs
                if getattr(obj, self.default_fk_stored_field_name) == response
            ]
            if not selection or qs.count() > 1:
                message = {m2m_field: error_msg or f"Expected {response} only."}
                self._errors.update(message)
                self._error_codes.append(M2M_SELECTION_ONLY)
                raise ValidationError(message, code=M2M_SELECTION_ONLY)
        return False

    def m2m_selections_not_expected(
        self,
        *responses: str | int | bool,
        m2m_field: str = None,
        error_msg: str | None = None,
    ) -> bool:
        """Raises an exception or returns False.

        m2m_field is required but no selections may be in `responses`.
        """
        qs = self.cleaned_data.get(m2m_field)
        if qs and qs.count() > 0:
            selections = [
                getattr(obj, self.default_fk_stored_field_name)
                for obj in qs
                if getattr(obj, self.default_fk_stored_field_name) in responses
            ]
            if selections:
                display_names = ", ".join(
                    [
                        getattr(obj, self.default_fk_display_field_name)
                        for obj in qs
                        if getattr(obj, self.default_fk_stored_field_name) in responses
                    ]
                )
                message = {
                    m2m_field: error_msg
                    or f"Invalid selection. Cannot be any of: {display_names}."
                }
                self._errors.update(message)
                self._error_codes.append(M2M_INVALID_SELECTION)
                raise ValidationError(message, code=M2M_INVALID_SELECTION)
        return False



from django import forms

from ...models import Confirmation


class ConfirmationForm(forms.ModelForm):
    class Meta:
        model = Confirmation
        fields = "__all__"
        help_text = {"confirmation_identifier": "(read-only)"}
        widgets = {
            "confirmation_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



import contextlib

from django.apps import AppConfig
from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, models


class LoadModelDataError(Exception):
    pass


def load_model_data(model_data: dict, apps: AppConfig | None = None) -> int:
    """Loads data into a model, creates or updates existing.

    Must have a unique field

    Format:
        {app_label.model1: [{field_name1: value,
                             field_name2: value ...},...],
         (app_label.model2, unique_field_name):
           [{field_name1: value,
             unique_field_name: value ...}, ...],
         ...}
    """
    apps = apps or django_apps
    n = 0
    for model_name, options in model_data.items():
        try:
            model_name, unique_field = model_name
        except ValueError:
            unique_field = None
        model = apps.get_model(model_name)
        unique_field = check_is_unique(model, unique_field) or guess_unique_field(model)
        for opts in options:
            try:
                obj = model.objects.get(**{unique_field: opts.get(unique_field)})
            except ObjectDoesNotExist:
                with contextlib.suppress(IntegrityError):
                    model.objects.create(**opts)
            else:
                for key, value in opts.items():
                    setattr(obj, key, value)
                obj.save()
        n += 1
    return n


def check_is_unique(model: models.Model, unique_field: str) -> str:
    if unique_field:
        fld_cls = getattr(model, unique_field)
        unique = fld_cls.field.unique
        if not unique:
            raise LoadModelDataError(f"Field is not unique. See {model}.{unique_field}")
    return unique_field


def guess_unique_field(model: models.Model) -> str:
    """Returns the first field name for a unique field."""
    unique_field = None
    for field in model._meta.get_fields():
        try:
            if field.unique and field.name != "id":
                unique_field = field.name
                break
        except AttributeError:
            pass
    if not unique_field:
        raise LoadModelDataError(
            f"Unable to determine unique field when loading model data. See {model}."
        )
    return unique_field
